import sys

import astropy.units as u
import numpy as np
from astropy.table import QTable
from astropy.table import Table
from astropy.table import vstack


def progressbar(it, prefix="", size=60, file=sys.stdout, label=""):
    count = len(it)
    if count > 0:

        def show(j):
            x = int(size * j / count)
            file.write(
                "%s[%s%s] %i/%i %s\r"
                % (prefix, "#" * x, "." * (size - x), j, count, label)
            )
            file.flush()

        show(0)
        for i, item in enumerate(it):
            yield item
            show(i + 1)
        file.write("\n")
        file.flush()


def to_dict(obj, classkey=None):
    if isinstance(obj, dict):
        data = {}
        for k, v in obj.items():
            data[k] = to_dict(v, classkey)
        return data
    elif isinstance(obj, u.Quantity):
        data = {"value": obj.value, "unit": str(obj.unit)}
        return data
    elif hasattr(obj, "_ast"):
        return to_dict(obj._ast())
    elif isinstance(obj, QTable):
        data = {"table": obj}
        #
        # commented lines convert table to dict, but table is already converted during the writing process!
        #
        # keys = [k for k in obj.keys()]
        # data_to_dict = {}
        # for k in keys:
        #     if hasattr(obj[k][0], 'unit'):
        #         data_to_dict[k] = {'value': obj[k], 'unit': str(obj[k][0].unit)}
        #     else:
        #         if isinstance(obj[k][0], str):
        #             data_to_dict[k] = {'value': [str(e) for e in obj[k]]}
        #         else:
        #             data_to_dict[k] = {'value': obj[k]}
        return data
    elif isinstance(obj, Table):
        keys = [k for k in obj.keys()]
        data = {}
        for k in keys:
            data[k] = {"value": obj[k]}
        return data

    elif hasattr(obj, "__dict__"):
        data = {
            key: to_dict(value, classkey)
            for key, value in obj.__dict__.items()
            if not callable(value)
            and not key.startswith("_")
            and key not in ["name"]
        }
        if classkey is not None and hasattr(obj, "__class__"):
            data[classkey] = obj.__class__.__name__
        return data
    else:
        return obj


def vstack_tables(table_list):
    table = False
    for tab in table_list:
        if not table:
            table = tab
        else:
            col_name1 = [col for col in table.keys() if col not in tab.keys()]
            if col_name1:
                for c in col_name1:
                    tab[c] = np.zeros(len(tab)) * table[c][0]
            col_name2 = [col for col in tab.keys() if col not in table.keys()]
            if col_name2:
                for c in col_name2:
                    table[c] = np.zeros(len(table)) * tab[c][0]
            table = vstack([table, tab], join_type="outer")
    return table


def parse_range(inp, avail_values):
    if inp.strip() == "all":
        return [x for x in range(avail_values)]
    if inp.strip() == "none":
        return []
    args = inp.split(",")

    numbers = []

    for a in args:
        start = None
        end = None
        try:
            s = a.split("-")
            start = int(s[0])
            end = int(s[1])
            end += 1
        except IndexError:
            start = int(a)
            end = int(a) + 1
        except ValueError:
            print("Invalid axis format, should be int")
            return None
        for x in range(start, end):
            numbers.append(x)
    return numbers


def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i : i + n]



from .task import Task


class EstimateNoiseInChannel(Task):
    """

    Parameters
    ----------
    target: Target
        target to prepare
    channel : dict
        channel to analyse
    payload : dict
        payload description

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")
        self.addTaskParam("channel", "channel to analyse")
        self.addTaskParam("payload", "payload common section")

    def execute(self):
        from exorad.models import noise
        import numpy as np
        import astropy.units as u
        from astropy.table import QTable

        channel = self.get_task_param("channel")
        target = self.get_task_param("target")
        payload = self.get_task_param("payload")

        name = channel["value"]
        self.info("estimating noise in {}".format(name))

        out = QTable()
        out = noise.frame_time(target, channel, out)
        read_gain, shot_gain = noise.multiaccum(channel, out["frameTime"][0])
        self.debug(
            "read gain : {} , shot gain : {}".format(read_gain, shot_gain)
        )

        out = noise.photon_noise(target.table, channel, shot_gain, out)

        out["darkcurrent_noise"] = (
            np.sqrt(
                shot_gain
                * target.table["WindowSize"][target.table["chName"] == name]
                * channel["detector"]["dark_current"]["value"]
                * u.ct
                / u.hr
            ).to(u.ct / u.s)
            * u.hr**0.5
        )
        self.debug("dark current noise : {}".format(out["darkcurrent_noise"]))
        out["read_noise"] = (
            np.sqrt(
                read_gain
                * channel["detector"]["read_noise"]["value"] ** 2
                * target.table["WindowSize"][target.table["chName"] == name]
                / out["frameTime"]
                * 1
                / u.hr
            ).to(u.ct / u.s)
            * u.hr**0.5
        )
        self.debug("read noise : {}".format(out["read_noise"]))

        if "NoiseX" in channel:
            NoiseX = channel["NoiseX"]["value"]
        else:
            NoiseX = 0.0
        self.debug("NoiseX: {}".format(NoiseX))

        photon_noise_variance = noise.photon_noise_variance(target.table, out)
        signal = target.table["star_signal_inAperture"][
            target.table["chName"] == name
        ].copy()
        signal[signal == 0.0] = np.nan
        out["total_noise"] = (
            np.sqrt(
                out["darkcurrent_noise"] ** 2
                + (1.0 + NoiseX) * (photon_noise_variance)
                + out["read_noise"] ** 2
            )
            / signal
        )

        wl = target.table["Wavelength"][target.table["chName"] == name]
        if payload:
            try:
                out = noise.add_custom_noise(
                    payload["common"]["customNoise"], wl, out
                )
            except KeyError:
                self.debug("no custom noise found in common section")
        try:
            out = noise.add_custom_noise(channel["customNoise"], wl, out)
        except KeyError:
            self.debug("no custom noise found in channel section")

        self.debug("total noise : {}".format(out["total_noise"]))

        self.set_output(out)


class EstimateNoise(Task):
    """

    Parameters
    ----------
    target: Target
        target to prepare
    channels : dict
        channel list to analyse

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam("target", "target to prepare")
        self.addTaskParam("channels", "channel list to analyse")

    def execute(self):
        from exorad.utils.util import vstack_tables
        from exorad.tasks import UpdateTargetTable
        from exorad.tasks import EstimateMaxSignal

        estimateMaxSignal = EstimateMaxSignal()
        estimateNoiseInChannel = EstimateNoiseInChannel()
        updateTargetTable = UpdateTargetTable()

        self.info("computing noise")
        target = self.get_task_param("target")
        channels = self.get_task_param("channels")

        target = estimateMaxSignal(target=target)

        table_list = []
        for ch in self.get_task_param("channels"):
            self.debug("computing noise in {}".format(ch))
            table_list.append(
                estimateNoiseInChannel(
                    target=target,
                    channel=channels[ch].description,
                    payload=channels[ch].payload,
                )
            )
        table = vstack_tables(table_list)
        new_target = updateTargetTable(target=target, table=table)
        self.set_output(new_target)



from collections import OrderedDict

import astropy.units as u
import numpy as np

from .task import Task
from exorad.models.foregrounds.skyForegrounds import SkyForeground
from exorad.models.foregrounds.zodiacalForeground import ZodiacalFrg
from exorad.utils.passVal import PassVal


class EstimateZodi(Task):
    """
    It estimate the zodiacal radiance in the target direction for a specific wl range

    Parameters
    -----------
    zodi: dict
        zodiacal foreground description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam("zodi", "zodiacal foreground description")
        self.addTaskParam("target", "target class")
        self.addTaskParam("wl_range", "wavelength range to investigate")

    def execute(self):
        self.info("estimating zodiacal foreground")
        zodi_dict = self.get_task_param("zodi")
        target = self.get_task_param("target")
        wl_min, wl_max = self.get_task_param("wl_range")

        wl = (
            np.logspace(
                np.log10((wl_min.to(u.um)).value),
                np.log10((wl_max.to(u.um)).value),
                PassVal.working_R,
            )
            * u.um
        )

        if "ra" in target.star.__dict__.keys():
            zodi = ZodiacalFrg(
                wl=wl,
                description=zodi_dict,
                coordinates=(
                    target.star.__dict__["ra"],
                    target.star.__dict__["dec"],
                ),
            )
        else:
            zodi = ZodiacalFrg(wl=wl, description=zodi_dict)

        if not hasattr(target, "foreground"):
            setattr(target, "foreground", OrderedDict())
        target.foreground["zodi"] = zodi.radiance
        self.set_output(target)


class EstimateForeground(Task):
    """
    It estimate the foreground radiance in the target direction for a specific wl range

    Parameters
    -----------
    foreground: dict
        foreground description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam("foreground", "foreground description")
        self.addTaskParam("target", "target class")
        self.addTaskParam("wl_range", "wavelength range to investigate")

    def execute(self):
        self.info("estimating custom foreground")
        foreground_dict = self.get_task_param("foreground")
        target = self.get_task_param("target")
        wl_min, wl_max = self.get_task_param("wl_range")

        wl = (
            np.logspace(
                np.log10((wl_min.to(u.um)).value),
                np.log10((wl_max.to(u.um)).value),
                PassVal.working_R,
            )
            * u.um
        )
        foreground_name = foreground_dict["value"]
        foreground = SkyForeground(wl, foreground_dict)
        if not hasattr(target, "foreground"):
            setattr(target, "foreground", OrderedDict())
        target.foreground[foreground_name] = foreground.skyFilter
        if not hasattr(target, "skyTransmission"):
            from exorad.models.signal import Signal

            setattr(
                target,
                "skyTransmission",
                Signal(wl, foreground.skyFilter.transmission),
            )
        else:
            target.skyTransmission.data *= foreground.skyFilter.transmission
        self.set_output(target)


class EstimateForegrounds(Task):
    """
    It estimate the foreground radiance in the target direction for a specific wl range

    Parameters
    -----------
    foregrounds: dict
        foregrounds description
    target: Target
        target class
    wl_range: (float, float)
        wavelength range to investigate. (wl_min, wl_max)

    Returns
    -------
    Target:
        updated target class
    """

    def __init__(self):
        self.addTaskParam("foregrounds", "foregrounds description")
        self.addTaskParam("target", "target class")
        self.addTaskParam("wl_range", "wavelength range to investigate")

    def execute(self):
        self.info("estimating foregrounds")
        target = self.get_task_param("target")
        foregrounds = self.get_task_param("foregrounds")
        wl_min, wl_max = self.get_task_param("wl_range")

        estimateZodi = EstimateZodi()
        estimateForeground = EstimateForeground()

        if isinstance(foregrounds, OrderedDict):
            for foreground in foregrounds:
                if foreground == "zodiacal":
                    target = estimateZodi(
                        zodi=foregrounds["zodiacal"],
                        target=target,
                        wl_range=(wl_min, wl_max),
                    )
                else:
                    target = estimateForeground(
                        foreground=foregrounds[foreground],
                        target=target,
                        wl_range=(wl_min, wl_max),
                    )
        else:
            if foregrounds["value"] == "zodiacal":
                target = estimateZodi(
                    zodi=foregrounds, target=target, wl_range=(wl_min, wl_max)
                )
            else:
                target = estimateForeground(
                    foreground=foregrounds,
                    target=target,
                    wl_range=(wl_min, wl_max),
                )
        self.set_output(target)



import logging
import os
import pathlib
import unittest

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from conf import skip_plot
from test_options import payload_file

from exorad.log import setLogLevel
from exorad.models.foregrounds.zodiacalForeground import ZodiacalFrg
from exorad.tasks import LoadTargetList
from exorad.tasks import PreparePayload
from exorad.tasks import PrepareTarget

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')

options_filename = payload_file()
target_list = os.path.join(data_dir, 'test_target.csv')


class ZodiacalTest(unittest.TestCase):
    wl = np.logspace(np.log10(0.45), np.log10(2.2), 6000) * u.um

    def test_radiance(self):
        setLogLevel(logging.DEBUG)

        payload = {'zodiacal': {'zodiacFactor': {'value': 2.5},
                                'zodicalFrgMap': False}}

        zodi = ZodiacalFrg(wl=self.wl, description=payload['zodiacal'])

        if not skip_plot:
            zodi.radiance.plot()
            plt.show()

    def test_fit_coordinate(self):
        setLogLevel(logging.DEBUG)

        payload = {'zodiacal': {'zodiacalMap': {'value':True}}}

        zodi = ZodiacalFrg(wl=self.wl, description=payload['zodiacal'],
                           coordinates=(90.03841366076144 * u.deg,
                                        -66.55432012293919 * u.deg))

        payload = {'zodiacal': {'zodiacFactor': {'value': 1.4536394185097168},
                                'zodiacalMap': {'value': False}}}

        zodi_validation = ZodiacalFrg(wl=self.wl,
                                      description=payload['zodiacal'])

        self.assertListEqual(list(zodi.radiance.data.value),
                             list(zodi_validation.radiance.data.value))


class BackgroundHanderTest(unittest.TestCase):
    setLogLevel(logging.INFO)

    preparePayload = PreparePayload()
    loadTargetList = LoadTargetList()
    prepareTarget = PrepareTarget()
    payload, channels, (wl_min, wl_max) = preparePayload(
        payload_file=options_filename, output=None)
    targets = loadTargetList(target_list=target_list)
    target = targets.target[0]

    target = prepareTarget(target=target, channels=channels)

    def test_handler(self):
        setLogLevel(logging.DEBUG)

        from exorad.tasks.foregroundHandler import EstimateZodi
        estimateBackground = EstimateZodi()
        target = estimateBackground(
            zodi=self.payload['common']['foreground']['zodiacal'],
            target=self.target,
            wl_range=(self.wl_min, self.wl_max))
        print(target.foreground['zodi'].data)



import logging
import os
import pathlib
import unittest

import h5py
from test_options import payload_file

from exorad.log import setLogLevel
from exorad.models.instruments import Photometer
from exorad.models.instruments import Spectrometer
from exorad.output.hdf5 import HDF5Output
from exorad.tasks import MergeChannelsOutput
from exorad.tasks.instrumentHandler import BuildChannels
from exorad.tasks.instrumentHandler import LoadPayload
from exorad.tasks.loadOptions import LoadOptions

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')
test_dir = os.path.join(path, 'test_data')

loadOptions = LoadOptions()
options = loadOptions(filename=payload_file())


class PhotometerTest(unittest.TestCase):
    setLogLevel(logging.DEBUG)
    photometer = Photometer('Phot', options['channel']['Phot'], options)
    photometer.build()

    def test_photometer_table(self):
        self.assertEqual(self.photometer.table['Wavelength'].value, 0.55)
        # self.assertEqual(self.photometer.table['TR'].value, 0.5)
        self.assertEqual(self.photometer.table['QE'].value, 0.55)


class SpectrometerTest(unittest.TestCase):
    setLogLevel(logging.DEBUG)
    spectrometer = Spectrometer('Spec', options['channel']['Spec'], options)
    spectrometer.build()

    def test_spectrometer_table(self):
        self.assertEqual(self.spectrometer.table['Wavelength'].size, 12)
        # self.assertListEqual(list(self.spectrometer.table['TR'].value), [0.5] * 12)
        self.assertListEqual(list(self.spectrometer.table['QE'].value),
                             [0.7] * 12)

    def test_spectrometer_native_R(self):
        from copy import deepcopy
        opt = deepcopy(options)
        opt['channel']['Spec']['targetR']['value'] = 'native'
        spectrometer = Spectrometer('Spec', opt['channel']['Spec'], opt)
        spectrometer.build()

    def test_spectrometer_no_R(self):
        from copy import deepcopy
        opt = deepcopy(options)
        opt['channel']['Spec'].pop('targetR')
        spectrometer = Spectrometer('Spec', opt['channel']['Spec'], opt)
        spectrometer.build()
        self.assertEqual(opt['channel']['Spec']['targetR']['value'], 'native')


class Spectrometer_from_pickle_Test(unittest.TestCase):
    setLogLevel(logging.DEBUG)
    loadOptions = LoadOptions()
    options2 = loadOptions(filename=
    payload_file(
        name='payload_test_pickle_spec.xml'))
    spectrometer = Spectrometer('Spec', options2['channel']['Spec'], options2)
    spectrometer.build()

    def test_spectrometer_table(self):
        self.assertEqual(self.spectrometer.table['Wavelength'].size, 12)
        # self.assertListEqual(list(self.spectrometer.table['TR'].value), [0.5] * 12)
        self.assertListEqual(list(self.spectrometer.table['QE'].value),
                             [0.7] * 12)


class InstrumentBuilderTest(unittest.TestCase):
    setLogLevel(logging.DEBUG)
    buildChannels = BuildChannels()
    channels = buildChannels(payload=options, write=False, output=None)

    def test_builder_dict(self):
        self.assertListEqual(list(self.channels.keys()), ['Phot', 'Spec'])


class IOTest(unittest.TestCase):
    setLogLevel(logging.INFO)

    buildChannels = BuildChannels()
    fname = 'test.h5'
    with HDF5Output(fname) as o:
        channels_built = buildChannels(payload=options, write=True, output=o)

    file = h5py.File(fname)
    loadPayload = LoadPayload()
    payload_loaded, channels_loaded = loadPayload(input=file)

    def test_instrument_list(self):
        self.assertListEqual(list(self.channels_built.keys()),
                             list(self.channels_loaded.keys()))

    def test_instrument_build_from_file(self):
        with self.assertRaises(ValueError):
            self.channels_loaded['Phot'].build()
        with self.assertRaises(ValueError):
            self.channels_loaded['Spec'].build()

        os.remove(self.fname)


class MergeOutputTest(unittest.TestCase):
    setLogLevel(logging.INFO)
    buildChannels = BuildChannels()
    mergeChannelsOutput = MergeChannelsOutput()
    channels = buildChannels(payload=options, write=False, output=None)
    setLogLevel(logging.DEBUG)

    def test_table_output(self):
        table = self.mergeChannelsOutput(channels=self.channels)



import glob
import os
import warnings

import numpy as np
from astropy import units as u
from astropy.io import ascii
from astropy.io import fits

from .signal import Sed
from exorad.log import Logger
from exorad.utils import exolib

warnings.filterwarnings("ignore", category=UserWarning, append=True)


class Star(Logger):
    """
    Instantiate a Stellar class using Phenix Stellar Models

    Attributes
    ----------
    lumiosity : 		float
                Stellar bolometric luminosity computed from Phenix stellar modle. Units [W]
    wl				array
                wavelength [micron]
    sed 			array
                Spectral energy density [W m**-2 micron**-1]
    ph_wl			array
                phoenix wavelength [micron]. Phenix native resolution
    ph_sed 			array
                phenix spectral energy density [W m**-2 micron**-1]. Phenix native resolution
    ph_filename 		phenix filename
    """

    # def __init__(self, star_sed_path, star_distance, star_temperature, star_logg, star_f_h, star_radius):
    def __init__(
        self,
        star_sed_path,
        starDistance,
        starTemperature,
        starLogg,
        starMetallicity,
        starRadius,
        use_planck_spectrum=False,
        wl_min=0.2 * u.um,
        wl_max=50.0 * u.um,
        phoenix_model_filename=None,
    ):
        """
        Parameters
        __________
          exocat_star 	: 	object
                    exodata star object
          star_sed_path:    : 	string
                    path to Phoenix stellar spectra

        """
        self.set_log_name()
        self.model = None

        if use_planck_spectrum == True:
            self.debug("Planck spectrum used")
            wl = np.linspace(wl_min, wl_max, 10000)
            ph_wl, ph_sed, ph_L = self.__get_star_spectrum(
                wl,
                starDistance.to(u.m),
                starTemperature.to(u.K),
                starRadius.to(u.m),
            )
            ph_file = None
            self.model = "Planck"
        else:
            if phoenix_model_filename:
                ph_file = os.path.join(star_sed_path, phoenix_model_filename)
                self.debug("phoenix file name : {}".format(ph_file))
            else:
                ph_file = self.__get_phonix_model_filename(
                    star_sed_path,
                    starTemperature.to(u.K),
                    starLogg,
                    starMetallicity,
                )
                self.debug("phoenix file name : {}".format(ph_file))

            ph_wl, ph_sed, ph_L = self.__read_phenix_spectrum(
                ph_file, starDistance.to(u.m), starRadius.to(u.m)
            )
            self.model = os.path.basename(ph_file)

        self.luminosity = ph_L
        self.sed = Sed(wl_grid=ph_wl, data=ph_sed)
        self.filename = ph_file

    def __get_sed_list(self, path):
        sed_name = []
        # todo include more phoenix formats
        format_list = [
            "*.BT-Settl.spec.fits.gz"
        ]  # , "*.7.bz2", "*.7.gz", "*HiRes.fits"]
        for format in format_list:
            sed_name = glob.glob(os.path.join(path, format))
            if len(sed_name) != 0:
                return sed_name

        if len(sed_name) == 0:
            self.error("No stellar SED files found")
            raise OSError("No stellar SED files found")

    def __get_phonix_model_filename(
        self, path, star_temperature, star_logg, star_f_h
    ):
        sed_name = self.__get_sed_list(path)
        sed_name_cleaned = [os.path.basename(k) for k in sed_name]

        sed_T_list = np.array(
            [float(name.split("-")[0][3:]) for name in sed_name_cleaned]
        )
        sed_Logg_list = np.array(
            [float(name.split("-")[1]) for name in sed_name_cleaned]
        )
        sed_Z_list = np.array(
            [float(name.split("-")[2][:3]) for name in sed_name_cleaned]
        )

        temp_to_find = star_temperature.value / 100
        # if 'HiRes' not in sed_name_cleaned[0]:
        #     temp_to_find /= 100.0

        if np.round(temp_to_find) < min(sed_T_list) or np.round(
            temp_to_find
        ) > max(sed_T_list):
            raise ValueError

        idx = np.argmin(
            np.abs(sed_T_list - np.round(temp_to_find))
            + np.abs(sed_Logg_list - star_logg)
            + np.abs(sed_Z_list - star_f_h)
        )

        ph_file = sed_name[idx]

        return ph_file

    def __read_phenix_spectrum(self, ph_file, star_distance, star_radius):
        """Read a PHENIX Stellar Spectrum.

        Parameters
        __________
          ph_file : 		string
                    full path to models file containing the SED
        Returns
        -------
          wl:			array
                    The Wavelength at which the SED is sampled. Units are [micron]
          sed :			array
                    The SED of the star. Units are [W m**-2 micron**-1]
          L :			scalar
                    The bolometric luminosity of the star. Units are [W]

        """

        ####################### USING PHOENIX BIN_SPECTRA BINARY FILES (h5)

        if "spec.fits.gz" in ph_file:
            with fits.open(ph_file) as hdu:
                strUnit = hdu[1].header["TUNIT1"]
                wl = hdu[1].data.field("Wavelength") * u.Unit(strUnit)

                strUnit = hdu[1].header["TUNIT2"]
                sed = hdu[1].data.field("Flux") * u.Unit(strUnit)

                bolometric_luminosity = (hdu[1].header["PHXLUM"] * u.W).to(
                    u.Lsun
                )

        # todo include more phoenix formats
        # elif 'HiRes.fits' in ph_file:
        #     import pathlib
        #     with fits.open(ph_file) as hdu:
        #         sed = hdu[0].data * u.erg/ u.s/ u.cm**2/ u.cm
        #         wl_ref = hdu[0].header['WAVE']
        #     path = pathlib.Path(ph_file).parent.absolute()
        #     wl_ref = wl_ref.replace('../../', '')
        #     wl_file = os.path.join(path, wl_ref)
        #     with fits.open(wl_file) as hdu:
        #         wl = hdu[0].data * u.cm
        #
        # elif '7.gz' in ph_file:
        #     import gzip
        #     with gzip.open(ph_file, 'rb') as f:
        #         lines = f.readlines()
        #         wl = [x.decode("utf-8") .split(' ')[2] for x in lines]
        #         sed = [x.decode("utf-8") .split(' ')[4] for x in lines]
        #
        # elif '7.bz2' in ph_file:
        #     import bz2
        #     with bz2.BZ2File(ph_file) as f:
        #         lines = f.readlines()
        #         wl, sed = [], []
        #         for line in lines:
        #             x = line.decode("utf-8").replace('D','E').split(' ')
        #             for i in range(len(x)):
        #                 if x[i] != '':
        #                     wl.append(float(x[i]))
        #                     break
        #             for j in range(len(x)):
        #                 if x[i+j] != '':
        #                     sed.append(float(x[i+j]))
        #                     break
        #     wl = np.array(wl) * u.angstrom
        #     sed = np.log10(np.array(sed)) * u.erg/ u.s/ u.cm**2/ u.angstrom
        #     idx = np.argsort(wl)
        #     wl = wl[idx]
        #     sed = sed[idx]

        else:
            raise OSError("unsupported PHOENIX format")

        # remove duplicates
        idx = np.nonzero(np.diff(wl))
        wl = wl[idx]
        sed = sed[idx]
        # Normalise SED to observed SED
        bolometric_flux = np.trapz(sed, x=wl)  # [W m**-2]
        bolometric_luminosity = (
            4 * np.pi * star_radius**2 * bolometric_flux
        )  # [W]
        sed *= (star_radius / star_distance) ** 2  # [W/m^2/mu]

        # When trading two lines below for two above you'll be using the phoenix
        # luminosity, but this invalidates current validation tables against exosim
        # norm = bolometric_luminosity.to(u.W) / (4.0*np.pi*star_distance**2)
        # sed = norm * sed / bolometric_flux

        return wl, sed, bolometric_luminosity.to(u.Lsun)

    def __get_star_spectrum(
        self, wl, star_distance, star_temperature, star_radius
    ):
        omega_star = np.pi * (star_radius / star_distance) ** 2 * u.sr
        sed = omega_star * exolib.planck(wl, star_temperature)
        bolometric_flux = np.trapz(sed, x=wl)
        bolometric_luminosity = (
            4.0 * np.pi * star_distance**2 * bolometric_flux
        )
        return wl, sed, bolometric_luminosity.to(u.Lsun)


class CustomSed(Logger):
    def __init__(self, fname, star_radius, star_distance):
        self.set_log_name()
        ph = ascii.read(fname, format="ecsv")
        ph_wl = ph["Wavelength"].data * ph["Wavelength"].unit
        ph_sed = ph["Sed"].data * ph["Sed"].unit

        bolometric_flux = np.trapz(ph_sed, x=ph_wl)  # [W m**-2]
        bolometric_luminosity = (
            4 * np.pi * (star_radius.to(u.m)) ** 2 * bolometric_flux
        )  # [W]
        ph_sed *= (
            star_radius.to(u.m) / star_distance.to(u.m)
        ) ** 2  # [W/m^2/mu]
        self.debug("custom sed used : {}".format(fname))

        self.sed = Sed(ph_wl, ph_sed)
        self.filename = fname
        self.luminosity = bolometric_luminosity.to(u.Lsun)
        self.model = "Custom"



import base64
import socket

import requests

from exorad import __url__
from exorad.__version__ import __version__ as current_ver
from exorad.log import Logger


class VersionError(Exception):
    pass


class VersionControl(Logger):
    def __init__(
        self, package_url=__url__, current_version=current_ver, force=False
    ):
        super().__init__()

        self.current_version = current_version
        if self._check_internet():
            self.content = self.get_content(package_url)
            if self.content:
                message, new_ver = self.get_message()

                if message != "":
                    self.warning("### NEW CODE VERSION ONLINE ###")
                    self.warning(
                        "your version: {}  - online version:{}".format(
                            self.current_version, new_ver
                        )
                    )
                    self.warning("changelog: \n {}".format(message))
                    if force:
                        raise VersionError(
                            "please update your code version. Your version: {} - online version:{}".format(
                                self.current_version, new_ver
                            )
                        )

    def get_content(self, package_url):
        splitted_url = package_url.split("/")
        url = (
            "https://api.github.com/repos/{}/{}/contents/CHANGELOG.md".format(
                splitted_url[-2], splitted_url[-1]
            )
        )
        self.debug("package url: {}".format(url))
        req = requests.get(url)
        if req.status_code == requests.codes.ok:
            self.status_code = True
            req = req.json()
            content = base64.b64decode(req["content"])
            content = content.decode("ascii")
            content = content.splitlines()
        else:
            self.status_code = False
            content = None
        return content

    def get_message(self):
        message = ""
        new_ver = ""
        for i, line in enumerate(self.content):
            if "## [" in line:
                if "## [Unreleased]" in line:
                    continue

                if "[{}]".format(self.current_version) in line:
                    self.debug("version already updated")
                    break
                else:
                    self.debug("found newer version online")
                    message += line + "\n"
                    new_ver = line[line.find("[") + 1 : line.find("]")]
                    for line_ in self.content[i + 1 :]:
                        if self.current_version in line_:
                            break
                        elif line_ != "":
                            message += line_ + "\n"
                    break
        return message, new_ver

    def _check_internet(self, host="8.8.8.8", port=53, timeout=3):
        """
        Host: 8.8.8.8 (google-public-dns-a.google.com)
        OpenPort: 53/tcp
        Service: domain (DNS/TCP)

        solution from https://stackoverflow.com/a/33117579
        """
        try:
            socket.setdefaulttimeout(timeout)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect(
                (host, port)
            )
            return True
        except OSError as ex:
            print(ex)
            return False



import logging
import os
import pathlib
import unittest

from test_options import payload_file

from exorad.log import setLogLevel
from exorad.tasks import EstimateZodi
from exorad.tasks import LoadSource
from exorad.tasks import LoadTargetList
from exorad.tasks import PreparePayload
from exorad.tasks import PrepareTarget
from exorad.tasks import PropagateForegroundLight
from exorad.tasks import PropagateTargetLight

preparePayload = PreparePayload()
loadTargetList = LoadTargetList()
loadSource = LoadSource()
prepareTarget = PrepareTarget()
estimateBackground = EstimateZodi()

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')

options_filename = payload_file()
target_list = os.path.join(data_dir, 'test_target.csv')


class TargetPropagationTest(unittest.TestCase):
    payload, channels, (wl_min, wl_max) = preparePayload(
        payload_file=options_filename, output=None)
    targets = loadTargetList(target_list=target_list)
    target = targets.target[0]
    target, sed = loadSource(target=target,
                             source=payload['common']['sourceSpectrum'],
                             wl_range=(wl_min, wl_max))
    target = prepareTarget(target=target, channels=channels)

    setLogLevel(logging.DEBUG)

    def test_target(self):
        propagateTargetLight = PropagateTargetLight()
        target = propagateTargetLight(channels=self.channels,
                                      target=self.target)

        print(target.table)


class BackgroundPropagationTest(unittest.TestCase):
    setLogLevel(logging.INFO)
    payload, channels, (wl_min, wl_max) = preparePayload(
        payload_file=options_filename, output=None)
    targets = loadTargetList(target_list=target_list)
    target = targets.target[0]

    target = estimateBackground(
        zodi=payload['common']['foreground']['zodiacal'],
        target=target,
        wl_range=(wl_min, wl_max))
    target = prepareTarget(target=target, channels=channels)

    setLogLevel(logging.DEBUG)

    def test_target(self):
        setLogLevel(logging.DEBUG)
        propagateBackgroundLight = PropagateForegroundLight()
        target = propagateBackgroundLight(channels=self.channels,
                                          target=self.target)

        print(target.table)



import glob
import logging
import os

import astropy.units as u
import mpmath
import numpy as np
import photutils
from astropy.io import fits
from scipy import signal
from scipy.integrate import cumtrapz
from scipy.interpolate import interp1d
from scipy.special import j1
from scipy.stats import binned_statistic

logger = logging.getLogger("exorad.exolib")


def rebin(x, xp, fp):
    """Resample a function fp(xp) over the new grid x, rebinning if necessary,
    otherwise interpolates
    Parameters
    ----------
    x	: 	array like
    New coordinates
    fp 	:	array like
    y-coordinates to be resampled
    xp 	:	array like
    x-coordinates at which fp are sampled

    Returns
    -------
    out	: 	array like
    new samples

    """

    if x.unit != xp.unit:
        logger.fatal(
            "Units mismatch in rebin {:s}, {:s}".format(x.unit, xp.unit)
        )
        raise ValueError

    idx = np.where(np.logical_and(xp > 0.9 * x.min(), xp < 1.1 * x.max()))[0]
    xp = xp[idx]
    fp = fp[idx]

    if not hasattr(fp, "unit"):
        logger.debug("No units found for fp. Forced to None")
        fp *= u.Unit()
    funits = fp.unit

    # remove NaNs
    id = np.where(np.isnan(xp))[0]
    if id.size > 0:
        logger.debug("Nans found in input x array: removing it")
        xp = np.delete(xp, id)
        fp = np.delete(fp, id)

    id = np.where(np.isnan(x))[0]
    if id.size > 0:
        logger.debug("Nans found in new x array: removing it")
        x = np.delete(x, id)

    # remove duplicates
    while np.diff(xp).min() == 0:
        logger.debug("duplicate found in input x array: removing it")
        id = np.argmin(np.diff(xp))
        xp = np.delete(xp, id)
        fp = np.delete(fp, id)

    while np.diff(x).min() == 0:
        logger.debug("duplicate found in new x array: removing it")
        id = np.argmin(np.diff(x))
        x = np.delete(x, id)

    if np.diff(xp).max() < np.diff(x).min():
        # Binning!
        logger.debug("binning")
        bin_x = 0.5 * (x[1:] + x[:-1])
        x0 = x[0] - (bin_x[0] - x[0]) / 2.0
        x1 = x[-1] + (x[-1] - bin_x[-1]) / 2.0
        bin_x = np.insert(bin_x, [0], x0)
        bin_x = np.append(bin_x, x1)
        new_f = (
            binned_statistic(xp, fp, bins=bin_x, statistic="mean")[0] * funits
        )

    #    idx = np.where(np.logical_and(xp >= x.min(), xp <= x.max()))[0]
    #    print np.trapz(fp[idx], x=xp[idx])
    #    print np.trapz(new_f, x=x)
    #    asd
    else:
        logger.debug("interpolating")

        # Interpolate !
        # new_f = np.interp(x, xp, fp, left=0.0, right=0.0)
        #         from scipy.interpolate import interp1d
        #         interpolator = interp1d(xp, fp, kind='cubic', fill_value=0.0, assume_sorted=False, bounds_error=False)
        #         new_f = interpolator(x)
        func = interp1d(
            xp,
            fp,
            fill_value=0.0,
            assume_sorted=False,
            bounds_error=False,
            kind="linear",
        )
        new_f = func(x)

        if not hasattr(new_f, "unit"):
            logger.debug("output units force to inputs fp")
            new_f *= funits

    return x, new_f


def rebin_(x, xp, fp):
    """Resample a function fp(xp) over the new grid x, rebinning if necessary,
    otherwise interpolates
    Parameters
    ----------
    x	: 	array like
    New coordinates
    fp 	:	array like
    y-coordinates to be resampled
    xp 	:	array like
    x-coordinates at which fp are sampled

    Returns
    -------
    out	: 	array like
    new samples

    """

    if x.unit != xp.unit:
        logger.fatal(
            "Units mismatch in rebin {:s}, {:s}".format(x.unit, xp.unit)
        )
        raise ValueError

    idx = np.where(np.logical_and(xp > 0.9 * x.min(), xp < 1.1 * x.max()))[0]
    xp = xp[idx]
    fp = fp[idx]

    if np.diff(xp).max() < np.diff(x).min():
        # Binning!
        c = cumtrapz(fp.value, x=xp.value) * fp.unit * xp.unit
        xpc = xp[1:]

        delta = np.gradient(x)
        new_c_1 = (
            np.interp(x - 0.5 * delta, xpc, c, left=0.0, right=0.0) * c.unit
        )
        new_c_2 = (
            np.interp(x + 0.5 * delta, xpc, c, left=0.0, right=0.0) * c.unit
        )
        new_f = (new_c_2 - new_c_1) / delta
    else:
        # Interpolate !
        new_f = np.interp(x, xp, fp, left=0.0, right=0.0) * fp.unit

    return x, new_f


def planck(wl, T):
    """Planck function.

    Parameters
    __________
      wl : 			array
                wavelength [micron]
      T : 			scalar
                Temperature [K]
                Spot temperature [K]
    Returns
    -------
      spectrum:			array
                The Planck spectrum  [W m^-2 sr^-1 micron^-1]
    """

    a = np.float64(1.191042768e8) * u.um**5 * u.W / u.m**2 / u.sr / u.um
    b = np.float64(14387.7516) * 1 * u.um * u.K
    try:
        x = b / (wl * T)
        bb = a / wl**5 / (np.exp(x) - 1.0)
    except ArithmeticError:
        bb = np.zeros_like(wl)
    return bb


def load_standard_psf(F_x, F_y, wl, delta_pix, hdr):
    k_x = delta_pix / (F_x * wl * hdr["CDELT2"])
    k_y = delta_pix / (F_y * wl * hdr["CDELT1"])
    xmin = -(hdr["CRPIX2"] - 1) * F_x * wl * hdr["CDELT2"]
    xmax = (hdr["NAXIS2"] - (hdr["CRPIX2"] - 1)) * F_x * wl * hdr["CDELT2"]
    ymin = -(hdr["CRPIX1"] - 1) * F_y * wl * hdr["CDELT1"]
    ymax = (hdr["NAXIS1"] - (hdr["CRPIX1"] - 1)) * F_y * wl * hdr["CDELT1"]
    extent = (xmin, xmax, ymin, ymax)
    return k_x, k_y, extent


def binnedPSF(
    F_x,
    F_y,
    wl,
    delta_pix,
    filename=None,
    format=None,
    PSFtype="AIRY",
    plot=False,
):
    if filename and os.path.exists(filename):
        if filename[-5:] == ".fits":
            with fits.open(os.path.expanduser(filename)) as hdu:
                ima = hdu[0].data
                hdr = hdu[0].header
                # define a kernel representing the detector pixel response
                # and use fractional pixel
                k_x, k_y, extent = load_standard_psf(
                    F_x, F_y, wl, delta_pix, hdr
                )
        elif os.path.isdir(filename):
            filenames = np.sort(glob.glob(filename + "*.fits"))
            psf_wl = []
            for file in filenames:
                with fits.open(os.path.expanduser(file)) as hdu:
                    psf_wl.append(hdu[0].header["WAVELEN"])
            idx = (np.abs(np.asarray(psf_wl) - wl.value)).argmin()
            with fits.open(os.path.expanduser(filenames[idx])) as hdu:
                ima = hdu[0].data
                hdr = hdu[0].header
                k_x, k_y, extent = load_standard_psf(
                    F_x, F_y, wl, delta_pix, hdr
                )
    else:
        x = np.linspace(-4.0, 4.0, 256)
        xx, yy = np.meshgrid(x, x)
        r = np.pi * np.sqrt(xx**2 + yy**2) + 1.0e-10

        ima = (2.0 * j1(r) / r) ** 2
        ima *= 0.25 * np.pi * (x[1] - x[0]) ** 2
        # print ima.sum()
        k_x = delta_pix / (F_x * wl * (x[1] - x[0]))
        k_y = delta_pix / (F_y * wl * (x[1] - x[0]))
        # print k_y, k_x
        extent = (
            -(ima.shape[1] // 2) * F_x * wl * (x[1] - x[0]),
            (ima.shape[1] // 2) * F_x * wl * (x[1] - x[0]),
            -(ima.shape[0] // 2) * F_y * wl * (x[1] - x[0]),
            (ima.shape[0] // 2) * F_y * wl * (x[1] - x[0]),
        )
        # normalise
        ima /= ima.sum()

    fk_x, ik_x = np.modf(k_x)
    fk_y, ik_y = np.modf(k_y)

    kernel = np.ones((int(ik_y) + 2, int(ik_x) + 2)) * fk_x.unit
    kernel[:, 0] *= 0.5 * fk_x
    kernel[:, -1] *= 0.5 * fk_x
    kernel[0, :] *= 0.5 * fk_y
    kernel[-1, :] *= 0.5 * fk_y

    imac = signal.convolve2d(ima, kernel, mode="same")

    if plot:
        plot_imac(imac, extent)

    return imac, kernel, extent


def load_pixel_psf_size(delta_pix, hdr):
    xmin = [-hdr["NAXIS2"] / 2.0 * delta_pix.value] * u.micron
    xmax = [hdr["NAXIS2"] / 2.0 * delta_pix.value] * u.micron
    ymin = [-hdr["NAXIS1"] / 2.0 * delta_pix.value] * u.micron
    ymax = [hdr["NAXIS1"] / 2.0 * delta_pix.value] * u.micron
    extent = (xmin, xmax, ymin, ymax)
    return extent


def pixel_based_psf(wl, delta_pix, filename):
    if filename[-5:] == ".fits":
        with fits.open(os.path.expanduser(filename)) as hdu:
            ima = hdu[0].data
            hdr = hdu[0].header
            extent = load_pixel_psf_size(delta_pix, hdr)

    elif os.path.isdir(filename):
        filenames = np.sort(glob.glob(filename + "*.fits"))
        psf_wl = []
        for file in filenames:
            with fits.open(os.path.expanduser(file)) as hdu:
                psf_wl.append(hdu[0].header["WAVELEN"])
        idx = (np.abs(np.asarray(psf_wl) - wl.value)).argmin()
        with fits.open(os.path.expanduser(filenames[idx])) as hdu:
            ima = hdu[0].data
            hdr = hdu[0].header
            extent = load_pixel_psf_size(delta_pix, hdr)

    return ima, np.array([1.0]), extent


def load_paos_psf(wl_group):
    from exorad.output.hdf5 import load

    scale = 1.0e6
    group = load(input_group=wl_group)

    img_key = list(group.keys())[-1]
    group = group[img_key]

    ima = group["amplitude"] ** 2
    dx = group["dx"] * u.micron * scale
    dy = group["dy"] * u.micron * scale
    extent = group["extent"] * u.micron * scale
    fratio = group["fratio"]

    return ima, dx, dy, extent, fratio


def interpolate_paos_psf(fd, wl):
    wavelengths = np.asarray(
        [key for key in list(fd.keys()) if key != "info"]
    ).astype(np.float64)
    wavelengths.sort()

    if len(wavelengths) == 1:
        logger.error("Can't interpolate, PAOS file has only one wavelength")
        raise ValueError(
            "Can't interpolate, PAOS file has only one wavelength"
        )

    idx0 = np.argmin(np.abs(wavelengths - wl))

    if np.min(wavelengths) <= wl <= np.max(wavelengths):
        idx1 = idx0 - 1 if wavelengths[idx0] - wl > 0 else idx0 + 1
    else:
        logger.warning(
            "Wavelength is outside interpolation region, extrapolating..."
        )
        idx1 = idx0 + 1 if wavelengths[idx0] - wl > 0 else idx0 - 1

    wl0, wl1 = wavelengths[idx0], wavelengths[idx1]

    ima0, dx0, dy0, extent0, fratio0 = load_paos_psf(wl_group=fd[f"{wl0}"])

    ima1, dx1, dy1, extent1, fratio1 = load_paos_psf(wl_group=fd[f"{wl1}"])

    dx0, dy0, extent0 = dx0.value, dy0.value, extent0.value
    dx1, dy1, extent1 = dx1.value, dy1.value, extent1.value

    ima = interp1d(
        x=[wl0, wl1],
        y=[ima0, ima1],
        kind="linear",
        fill_value="extrapolate",
        axis=0,
    )(wl)
    dx = interp1d(
        x=[wl0, wl1], y=[dx0, dx1], kind="linear", fill_value="extrapolate"
    )(wl)
    dy = interp1d(
        x=[wl0, wl1], y=[dy0, dy1], kind="linear", fill_value="extrapolate"
    )(wl)
    extent = interp1d(
        x=[wl0, wl1],
        y=[extent0, extent1],
        kind="linear",
        fill_value="extrapolate",
        axis=0,
    )(wl)
    fratio = interp1d(
        x=[wl0, wl1],
        y=[fratio0, fratio1],
        kind="linear",
        fill_value="extrapolate",
    )(wl)

    return ima, dx * u.micron, dy * u.micron, extent * u.micron, fratio


def paosPSF(wl, delta_pix, filename="", plot=False):
    import h5py
    from scipy import signal

    logger.debug("loading PAOS psf")

    assert wl.unit == u.micron, print("Wavelength unit should be micron. ")

    try:
        with h5py.File(os.path.expanduser(filename), mode="r") as fd:
            wavelength = wl.value[0]
            if str(wavelength) in list(fd.keys()):
                ima, dx, dy, extent, fratio = load_paos_psf(
                    wl_group=fd[f"{wavelength}"]
                )
            else:
                ima, dx, dy, extent, fratio = interpolate_paos_psf(
                    fd, wavelength
                )

    except OSError as e:
        logger.error("Error loading PAOS psf file. ")
        raise FileNotFoundError(e.errno)

    k_x = delta_pix / dx
    k_y = delta_pix / dy

    fk_x, ik_x = np.modf(k_x)
    fk_y, ik_y = np.modf(k_y)

    kernel = np.ones((int(ik_y) + 2, int(ik_x) + 2)) * fk_x.unit

    kernel[:, 0] *= 0.5 * fk_x
    kernel[:, -1] *= 0.5 * fk_x
    kernel[0, :] *= 0.5 * fk_y
    kernel[-1, :] *= 0.5 * fk_y

    imac = signal.convolve2d(ima, kernel, mode="same")

    if plot:
        plot_imac(imac, extent)

    return imac, kernel, extent


def plot_imac(imac, extent, xlim=None, ylim=None):
    import matplotlib.pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable

    extent = np.asarray([ext.value for ext in extent]).ravel()

    fig, ax = plt.subplots(1, 1, figsize=(8, 6))

    im = ax.imshow(imac, cmap=plt.get_cmap("viridis"), origin="lower")
    extent = np.asarray(extent).astype(np.float64)
    im.set_extent(extent=extent)

    if (xlim, ylim) != (None, None):
        plt.xlim(*xlim)
        plt.ylim(*ylim)

    plt.grid(True)

    cax = make_axes_locatable(ax).append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im, cax=cax, orientation="vertical")

    plt.show()

    return 0


def wl_encircled_energy(filename, eec, format, Fnum_x, Fnum_y, delta_pix):
    filenames = np.sort(glob.glob(filename + "*.fits"))
    psf_wl, enc = [], []
    for file in filenames:
        with fits.open(os.path.expanduser(file)) as hdu:
            psf_wl.append(hdu[0].header["WAVELEN"])
        if format == "pixel_based":
            ima, _, _ = pixel_based_psf(
                hdu[0].header["WAVELEN"], delta_pix, file
            )
            enc += [find_aperture_radius(ima, eec, 1, 1, 1)]
        else:
            ima, _, _ = binnedPSF(
                Fnum_x, Fnum_y, hdu[0].header["WAVELEN"], delta_pix
            )
            enc += [
                find_aperture_radius(
                    ima, eec, Fnum_x, Fnum_y, hdu[0].header["WAVELEN"]
                )
            ]
    return psf_wl, enc


def find_aperture_radius(ima, eec, Fnum_x, Fnum_y, wavelength):
    """
    It finds the aperture radius for a given PSF such that the desired Encircled Energy is contained.

    Parameters
    ----------
    ima:
        psf image in micron scale
    eec:
        desired encircled energy
    Fnum_x:
        f number in the spectral direction
    Fnum_y:
        f number in the spatial direction
    wavelength:
        wavelength of the sampled psf

    Returns
    ---------
    float
    """
    last_enc, i = 0, 0
    rs, enc = [], []
    while last_enc < (eec + 0.02):
        i += 1
        rs += [i]
        enc += [encircled_energy(i, ima, Fnum_x, Fnum_y, wavelength)]
        last_enc = enc[-1]
    inter = interp1d(enc, rs)
    r = inter(eec)
    return r


def encircled_energy(r, ima, Fnum_x, Fnum_y, wavelength):
    r_pix_x = r * Fnum_x * wavelength
    r_pix_y = r * Fnum_y * wavelength
    aper = photutils.aperture.EllipticalAperture(
        (ima.shape[1] // 2, ima.shape[0] // 2),
        b=float(r_pix_y.value),
        a=float(r_pix_x.value),
    )
    phot_ = photutils.aperture.aperture_photometry(ima, aper)
    phot = phot_["aperture_sum"].data[0]

    return phot / ima.sum()


def OmegaPix(Fnum_x, Fnum_y=None):
    """
    Calculate the solid angle subtended by an elliptical aperture on-axis.
    Algorithm from "John T. Conway. Nuclear Instruments and Methods in
    Physics Research Section A: Accelerators, Spectrometers, Detectors and
    Associated Equipment, 614(1):17 ? 27, 2010.
    https://doi.org/10.1016/j.nima.2009.11.075
    Equation n. 56

    Parameters
    ----------
    Fnum_x : scalar
             Input F-number along dispersion direction
    Fnum_y : scalar
             Optional, input F-number along cross-dispersion direction

    Returns
    -------
     Omega : scalar
             The solid angle subtanded by an elliptical aperture in units u.sr

    """

    if not Fnum_y:
        Fnum_y = Fnum_x

    if Fnum_x > Fnum_y:
        a = 1.0 / (2 * Fnum_y)
        b = 1.0 / (2 * Fnum_x)
    else:
        a = 1.0 / (2 * Fnum_x)
        b = 1.0 / (2 * Fnum_y)

    h = 1.0

    A = 4.0 * h * b / (a * np.sqrt(h**2 + a**2))
    k = np.sqrt((a**2 - b**2) / (h**2 + a**2))
    alpha = np.sqrt(1 - (b / a) ** 2)

    Omega = 2.0 * np.pi - A * mpmath.ellippi(alpha**2, k**2)

    return Omega * u.sr


if __name__ == "__main__":
    Omega = OmegaPix(5.0, 5.0)

    sigma = np.arctan(1.0 / 10)
    OmegaCheck = 2.0 * np.pi * (1 - np.cos(sigma))
    print((Omega, OmegaCheck))



import logging
import os
import pathlib
import unittest

from exorad.log import setLogLevel
from exorad.tasks import GetChannelList
from exorad.tasks.loadOptions import LoadOptions

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')
test_dir = os.path.join(path, 'test_data')
setLogLevel(logging.DEBUG)


def payload_file(source=path.parent.absolute(), destination=test_dir,
                 name='payload_test.xml'):
    payload_file_ = os.path.join(data_dir, 'payload_example.xml')
    new_configPath = '    <ConfigPath> {}\n'.format(source)
    tmp = os.path.join(destination, name)
    try:
        os.remove(tmp)
    except OSError:
        pass
    with open(tmp, 'w+') as new_file:
        with open(payload_file_) as old_file:
            for line in old_file:
                if '<ConfigPath>' in line:
                    new_file.write(new_configPath)
                else:
                    new_file.write(line)
    return tmp


class LoadOptionsTest(unittest.TestCase):
    loadOptions = LoadOptions()

    def test_loadFile(self):
        self.loadOptions(filename=payload_file())
        with self.assertRaises(IOError): self.loadOptions(
            filename=os.path.join(test_dir,
                                  'payload_example_missing_data_file.xml'))
        with self.assertRaises(IOError): self.loadOptions(
            filename=os.path.join(path, 'test_data/payload_example.csv'))
        with self.assertRaises(IOError): self.loadOptions(
            filename=os.path.join(path, 'test_data/payload_example1.xml'))

        self.loadOptions(
            filename=os.path.join(test_dir,
                                  'payload_example_missing_config.xml'),
            config_path=str(path.parent.absolute()))

    def test_loadpickle(self):
        if not os.path.exists(os.path.join(test_dir, 'spec.pickle')):
            options = self.loadOptions(filename=payload_file())
            import pickle
            filehandler = open(os.path.join(test_dir, 'spec.pickle'), 'wb')
            pickle.dump(options['channel']['Spec'], filehandler)
            filehandler.close()
        options2 = self.loadOptions(filename=
                                    payload_file(
                                        name='payload_test_pickle_spec.xml'))


class GetChannelListTest(unittest.TestCase):
    options = {'channel': {'Phot': {'channelClass': {'value': 'Photometer'}},
                           'Spec': {'channelClass': {'value': 'Spectrometer'}},
                           'Spec1': {'channelClass': {'value': 'Spectrometer'}}
                           }
               }

    getChannelList = GetChannelList()

    def test_channel_types(self):
        photometer_list = self.getChannelList(options=self.options,
                                              channel_type='Photometer')
        self.assertListEqual(photometer_list, ['Phot'])
        photometer_list = self.getChannelList(options=self.options,
                                              channel_type='Spectrometer')
        self.assertListEqual(photometer_list, ['Spec', 'Spec1'])
        photometer_list = self.getChannelList(options=self.options,
                                              channel_type='Bolometer')
        self.assertListEqual(photometer_list, [])



from collections import OrderedDict

from astropy.io.misc.hdf5 import read_table_hdf5

from .task import Task
from exorad.models.instruments import Photometer
from exorad.models.instruments import Spectrometer
from exorad.output.hdf5 import load

instruments = {"photometer": Photometer, "spectrometer": Spectrometer}


class BuildInstrument(Task):
    """
    Initialize and build an instrument

    Parameters
    ----------
    type: str
        instrument class
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None
    write: bool
        set to True to write the built dict to file. Default is None
    output: str
        output object

    Returns
    -------
    Instrument:
        return the built instrument class

    Raises
    ------
        KeyError
            if the indicated instrument class is not supported

    Examples
    --------
    >>> buildInstrument = BuildInstrument()
    >>> instrument = buildInstrument(type='photometer', name='Phot',
    >>>                              description=payload['channel']['Phot'],
    >>>                              payload=payload, write=False, output=None)
    """

    def __init__(self):
        self.addTaskParam("type", "instrument type")
        self.addTaskParam("name", "instrument name")
        self.addTaskParam("description", "instrument description dictionary")
        self.addTaskParam("payload", "main payload. Default is None")
        self.addTaskParam("write", "write processed instrument to output file")
        self.addTaskParam("output", "output object")

    def execute(self):
        try:
            instrumentClass = instruments[self.get_task_param("type")]
        except KeyError:
            self.error("invalid instrument class")
            raise ValueError
        instrument = instrumentClass(
            self.get_task_param("name"),
            self.get_task_param("description"),
            self.get_task_param("payload"),
        )
        instrument.build()
        if self.get_task_param("write"):
            instrument.write(self.get_task_param("output"))
        self.set_output(instrument)


class BuildChannels(Task):
    """
    Initialize and build all the channels in the payload

    Parameters
    ----------
    payload: dict
        main payload. Default is None
    write: bool
        set to True to write the built dict to file. Default is None
    output: str
        output object
    Returns
    -------
    dict:
        return a dict of built Instrument classes

    Examples
    --------
    >>> buildChannels = BuildChannels()
    >>> channels = buildChannels(payload=payload, write=False, output=None)
    """

    def __init__(self):
        self.addTaskParam("payload", "main payload")
        self.addTaskParam("write", "write processed instrument to output file")
        self.addTaskParam("output", "output object")

    def execute(self):
        self.info("building channel")
        channels = {}
        self.debug(
            "detectors found : {}".format(
                self.get_task_param("payload")["channel"].keys()
            )
        )

        ch = None
        if self.get_task_param("write"):
            inst = self.get_task_param("output").create_group("payload")
            inst.store_dictionary(
                self.get_task_param("payload"),
                group_name="payload description",
            )
            ch = inst.create_group("channels")

        buildInstrument = BuildInstrument()
        if isinstance(self.get_task_param("payload")["channel"], OrderedDict):
            for det in self.get_task_param("payload")["channel"].keys():
                channel_type = self.get_task_param("payload")["channel"][det][
                    "channelClass"
                ]["value"].lower()
                channels[det] = buildInstrument(
                    type=channel_type,
                    name=det,
                    description=self.get_task_param("payload")["channel"][det],
                    payload=self.get_task_param("payload"),
                    write=False,
                    output=None,
                )
                if self.get_task_param("write"):
                    channels[det].write(ch)
        else:
            channel_type = self.get_task_param("payload")["channel"][
                "channelClass"
            ]["value"].lower()
            det = self.get_task_param("payload")["channel"]["value"]
            channels[det] = buildInstrument(
                type=channel_type,
                name=det,
                description=self.get_task_param("payload")["channel"],
                payload=self.get_task_param("payload"),
                write=False,
                output=None,
            )
            if self.get_task_param("write"):
                channels[det].write(ch)
        self.debug("channels : {}".format(channels))
        self.set_output(channels)


class LoadPayload(Task):
    """
    Loads payload and channels from dict

    Parameters
    ----------
    input: dict
        main dictionary


    Returns
    -------
    dict:
        payload dict
    dict:
        return a dict of built Instrument classes

    Examples
    --------
    >>> loadPayload = LoadPayload()
    >>> channels = loadPayload(input=input)
    """

    def __init__(self):
        self.addTaskParam("input", "input data")

    def execute(self):
        payload_dir = self.get_task_param("input")["payload"]
        payload = load(payload_dir["payload description"])
        channels_dir = payload_dir["channels"]
        channels = {}
        for ch in channels_dir.keys():
            ch_dir = channels_dir[ch]
            description = load(ch_dir["description"])
            instrument = instruments[
                description["channelClass"]["value"].lower().decode("utf-8")
            ]
            channels[ch] = instrument(
                name=ch,
                description=description,
                payload=payload,
            )
            table = read_table_hdf5(ch_dir, path=ch)
            built_instr = load(ch_dir["built_instr"])
            channels[ch].load(table, built_instr)
        self.debug("channels loaded: {}".format(channels))
        self.set_output([payload, channels])


class PreparePayload(Task):
    """
    It loads the payload information and returns a dictionary containing the payload,
    a dictionary with the channels already built,
    and a couple with minimum and maximum wavelength investigated by the payload

    Parameters
    ----------
    payload_file: str
        xml file with payload description
    output: str
        h5 output file

    Returns
    -------

    dict:
        payload dictionary
    dict:
        channels dictionary
    couple:
        minimum and maximum investigated wavelength

    Examples
    --------
    >>> preparePayload = PreparePayload()
    >>> payload, channels, (wl_min, wl_max) = PreparePayload(payload_file=payload, output=output)

    """

    def __init__(self):
        self.addTaskParam("payload_file", "payload xml file")
        self.addTaskParam("output", "output file")

    def execute(self):
        import os
        from exorad.tasks import LoadOptions
        from exorad.output.hdf5 import HDF5Output

        loadOptions = LoadOptions()
        buildChannels = BuildChannels()
        loadPayload = LoadPayload()

        payload_file = self.get_task_param("payload_file")
        output = self.get_task_param("output")

        if isinstance(payload_file, str):
            ext = os.path.splitext(payload_file)[1]
        else:
            ext = None

        if (ext == ".xml") or isinstance(payload_file, dict):
            if ext == ".xml":
                payload = loadOptions(filename=payload_file)
            if isinstance(payload_file, dict):
                payload = payload_file
            if output is not None:
                if os.path.isfile(output):
                    append = True
                else:
                    append = False
                with HDF5Output(output, append=append) as out:
                    channels = buildChannels(
                        payload=payload, write=True, output=out
                    )
            else:
                channels = buildChannels(
                    payload=payload, write=False, output=None
                )

        elif ext in [".h5", ".hdf5"]:
            import h5py

            with h5py.File(payload_file, "r") as f:
                payload, channels = loadPayload(input=f)
        #            payload, channels = loadPayload(input=payload_file)
        else:
            self.error("Unsupported payload format")
            raise OSError("Unsupported payload format")
        wl_min, wl_max = (
            payload["common"]["wl_min"]["value"],
            payload["common"]["wl_max"]["value"],
        )
        self.set_output([payload, channels, (wl_min, wl_max)])


class MergeChannelsOutput(Task):
    """
    Merges the channels output tables

    Parameters
    ----------
    channels: dict

    Returns
    ---------
    Table
    """

    def __init__(self):
        self.addTaskParam("channels", "dict of channels")

    def execute(self):
        from exorad.utils.util import vstack_tables

        channels = self.get_task_param("channels")
        table_list = []
        for channel in channels.keys():
            table_list.append(channels[channel].table)
        table = vstack_tables(table_list)
        self.set_output(table)


class GetChannelList(Task):
    """
    Returns a list of the channels of the specific type that are contained in the instrument options file

    Parameters
    ----------
    options: object
        payload description
    channel_type: string
        type of channel to find

    Returns
    -------
    list:
        list of strings with channel names

    Examples
    --------
    >>> from exorad.tasks.loadOptions import LoadOptions
    >>> loadOptions = LoadOptions()
    >>> options = loadOptions(filename = 'path/to/file.xml')
    >>> getDetectorList = GetChannelList()
    >>> photometers_list = getDetectorList(options = options, channel_type = 'Photometer')
    >>> spectrometers_list = getDetectorList(options = options, channel_type = 'Spectrometer')
    """

    def __init__(self):
        self.addTaskParam("options", "instrument description class")
        self.addTaskParam("channel_type", "desired channel type")

    def execute(self):
        self._opt = self.get_task_param("options")
        self._channel_type = self.get_task_param("channel_type")
        self.channel_lst = self.__get_channel_list__()
        self.info(
            "{} type channel found: {}".format(
                self._channel_type, self.channel_lst
            )
        )
        self.set_output(self.channel_lst)

    def __get_channel_list__(self):
        channel_lst = []
        for channel in self._opt["channel"].keys():
            if (
                self._opt["channel"][channel]["channelClass"]["value"].lower()
                == self._channel_type.lower()
            ):
                channel_lst.append(channel)
                self.debug(
                    "{} added as {}".format(channel, self._channel_type)
                )
        return channel_lst



import os

import numpy as np
from astropy import units as u

from exorad.models.source import CustomSed
from exorad.models.source import Star
from exorad.tasks.task import Task
from exorad.utils.passVal import PassVal


class LoadSource(Task):
    """
    Updates target with its source and return the source Sed

    Parameters
    ----------
    target: Target Class
        target class
    source: dict
        source spectrum description
    wl_range: couple
        wavelength range to investigate: (wl_min, wl_max)


    Returns
    -------
        Target:
            Target class with star.sed updated
        Sed:
            source Sed

    Examples
    --------
    >>> loadSource = LoadSource()
    >>> target, source = loadSource(target= target, source={'sourceSpectrum': {'value':'Planck'}})

    Raises
    ------
    AttributeError:
        if some target information are missing
    """

    def __init__(self):
        self.addTaskParam("target", "target class object")
        self.addTaskParam("source", "source spectrum description")
        self.addTaskParam("wl_range", "wavelength range to investigate")

    def execute(self):
        target = self.get_task_param("target")
        source = self.get_task_param("source")

        if isinstance(source, str):
            source = {"value": source}
            self.warning("source should be dict, not string.")

        wl_min, wl_max = self.get_task_param("wl_range")
        if wl_min > wl_max:
            wl_min, wl_max = wl_max, wl_min
        if not hasattr(wl_min, "unit"):
            self.debug("wavelength unit not found: micron assumed")
            wl_min *= u.um
            wl_max *= u.um

        wl_grid = (
            np.logspace(
                np.log10(wl_min.value),
                np.log10(wl_max.value),
                PassVal.working_R,
            )
            * wl_max.unit
        )

        if source["value"].lower() == "custom":
            # if custom source, only R and D are needed for the solid angle
            for attr in ["D", "R"]:
                if not hasattr(target.star.__getattribute__(attr), "value"):
                    self.error("target information incomplete")
                    raise AttributeError("target information incomplete")
            star = CustomSed(
                source["CustomSed"]["value"], target.star.R, target.star.D
            )
            self.debug(
                "custom sed used {}".format(source["CustomSed"]["value"])
            )

        else:
            # check if star information are complete
            for attr in ["D", "Teff", "M", "R"]:
                if not hasattr(target.star.__getattribute__(attr), "value"):
                    self.error("target information incomplete")
                    raise AttributeError("target information incomplete")

            self.debug("source spectrum : {}".format(source["value"].lower()))
            if source["value"].lower() == "planck":
                self.debug("Plack sed selected")
                star = Star(
                    star_sed_path=".",
                    starDistance=target.star.D,
                    starTemperature=target.star.Teff,
                    starLogg=target.star.calc_logg(
                        target.star.M, target.star.R
                    ),
                    starMetallicity=0.0,
                    starRadius=target.star.R,
                    use_planck_spectrum=True,
                    wl_min=wl_min,
                    wl_max=wl_max,
                )

            elif source["value"].lower() == "phoenix":
                try:
                    star_sed_path = source["StellarModels"]["value"]
                except KeyError:
                    if os.environ.get("PHOENIX_PATH", None) is not None:
                        star_sed_path = os.environ.get("PHOENIX_PATH", None)
                    else:
                        raise OSError("No phoenix path specificed")

                if not os.path.exists(star_sed_path):
                    raise OSError(
                        "Phoenix path does not exist: {}".format(star_sed_path)
                    )

                try:
                    star = Star(
                        star_sed_path=star_sed_path,
                        starDistance=target.star.D,
                        starTemperature=target.star.Teff,
                        starLogg=target.star.calc_logg(
                            target.star.M, target.star.R
                        ),
                        starMetallicity=0.0,
                        starRadius=target.star.R,
                        use_planck_spectrum=False,
                        wl_min=wl_min,
                        wl_max=wl_max,
                    )
                    self.debug("stellar sed used {}".format(star.filename))
                except ValueError:
                    self.warning(
                        "stellar temperature out sed boundaries: Planck star used instead"
                    )
                    star = Star(
                        star_sed_path=".",
                        starDistance=target.star.D,
                        starTemperature=target.star.Teff,
                        starLogg=target.star.calc_logg(
                            target.star.M, target.star.R
                        ),
                        starMetallicity=0.0,
                        starRadius=target.star.R,
                        use_planck_spectrum=True,
                        wl_min=wl_min,
                        wl_max=wl_max,
                    )
            else:
                star = Star(
                    star_sed_path=".",
                    starDistance=target.star.D,
                    starTemperature=target.star.Teff,
                    starLogg=target.star.calc_logg(
                        target.star.M, target.star.R
                    ),
                    starMetallicity=0.0,
                    starRadius=target.star.R,
                    use_planck_spectrum=True,
                    wl_min=wl_min,
                    wl_max=wl_max,
                )
                self.info(
                    "invalid source spectrum description. Planck spectrum is used"
                )

        star.sed.spectral_rebin(wl_grid)

        target.update_target(star)
        if hasattr(target, "table"):
            target.table = self._add_star_metadata(target.star, target.table)
        self.set_output([target, star.sed])

    def _add_star_metadata(self, star, table):
        metadata = {}

        metadata["starName"] = star.name
        metadata["starM"] = star.M
        metadata["starTeff"] = star.Teff
        metadata["starR"] = star.R
        metadata["starDistance"] = star.D
        metadata["starL"] = star.luminosity
        metadata["starModel"] = star.model
        if hasattr(star, "magk"):
            metadata["starMagK"] = star.magk

        table.meta.update(metadata)
        return table



import numpy as np
from astropy import constants as cc
from astropy import units as u

from exorad.log.logger import Logger
from exorad.models.source import CustomSed
from exorad.models.source import Star


class Target(Logger):
    """
    Target base class
    """

    def __init__(self):
        self.set_log_name()
        luminosity = None
        sed = None
        model = None
        table = None
        id = None
        star = None
        planet = None
        name = None

    def calc_logg(self, M, R):
        m = M.si
        r = R.si
        g = cc.G * m / r**2
        g = g.to(u.cm / u.s**2)

        return np.log10(g.value)

    def update_target(self, obj):
        if isinstance(obj, Star) or isinstance(obj, CustomSed):
            self.star.luminosity = obj.luminosity
            self.star.sed = obj.sed
            self.star.model = obj.model
            self.debug("target updated")
        else:
            self.warning("Object type {:s} not implemented".format(type(obj)))

    def write(self, output):
        targets_out = output.create_group("targets")
        group_name = str(self.name)
        target_dict = self.to_dict()
        targets_out.store_dictionary(target_dict, group_name=group_name)
        self.info("target {} saved".format(self.name))
        return output

    def to_dict(self):
        from exorad.utils.util import to_dict

        return to_dict(self)



import logging
import os
import pathlib
import shutil

import matplotlib.pyplot as plt

import exorad.__version__ as version
import exorad.tasks as tasks
from exorad.log import addLogFile
from exorad.log import setLogLevel
from exorad.output.hdf5 import HDF5Output
from exorad.utils.plotter import Plotter

logger = logging.getLogger("exorad")

preparePayload = tasks.PreparePayload()
mergeChannelsOutput = tasks.MergeChannelsOutput()
loadTargetList = tasks.LoadTargetList()
loadSource = tasks.LoadSource()
observeTarget = tasks.ObserveTarget()
observeTargetList = tasks.ObserveTargetlist()


def efficiency_plot(channels, output_dir=None):
    table = mergeChannelsOutput(channels=channels)
    plotter = Plotter(channels=channels, input_table=table)
    plotter.plot_efficiency()
    if output_dir is None:
        plt.show()
    else:
        plotter.save_fig(
            os.path.join(output_dir, "efficiency.png"), efficiency=True
        )
        logger.info("efficiency plot saved in output directory")
    plt.close()


def standard_pipeline(
    options,
    target_list,
    output=None,
    plot=False,
    full_contrib=False,
    n_thread=1,
    debug=False,
    log=False,
    replace=True,
):
    from exorad.utils.ascii_art import ascii_art

    logger.info(ascii_art)
    logger.info("code version {}".format(version.__version__))

    if debug:
        setLogLevel(logging.DEBUG)
    if isinstance(log, str):
        addLogFile(fname=log)
    elif log:
        addLogFile()

    if output is not None:
        if replace:
            if os.path.exists(output):
                os.remove(output)

        out_dir = pathlib.Path(output).parent.absolute()
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
            logger.info("output directory created")
        logger.info("output directory set as {}".format(out_dir))
        try:
            shutil.copy(options, out_dir)
        except shutil.SameFileError:
            pass
        try:
            shutil.copy(target_list, out_dir)
        except shutil.SameFileError:
            pass
    else:
        out_dir = None

    # step 1 load payload
    payload, channels, (wl_min, wl_max) = preparePayload(
        payload_file=options, output=output
    )
    if full_contrib:
        from astropy.table import hstack

        for ch in channels:
            channels[ch].table = hstack(
                [
                    channels[ch].table,
                    channels[ch].built_instr["optical_path"]["signal_table"],
                ]
            )
            channels[ch].table.remove_column("instrument_signal")

    # step 1b plot payload efficiency
    if plot:
        efficiency_plot(channels=channels, output_dir=out_dir)

    # step 2 load targetlist
    targets = loadTargetList(target_list=target_list)

    # step 3 observe targetlist
    targets = observeTargetList(
        targets=targets.target,
        payload=payload,
        channels=channels,
        wl_range=(wl_min, wl_max),
        plot=plot,
        out_dir=out_dir,
        n_thread=n_thread,
        debug=debug,
    )
    # step 4 save to output
    if output is not None:
        for target in targets:
            with HDF5Output(output, append=True) as out:
                targets[target].write(out)


def main():
    import argparse

    parser = argparse.ArgumentParser(description="ExoRad {}".format(version))
    parser.add_argument(
        "-p",
        "--payload",
        dest="opt",
        type=str,
        required=True,
        help="Input payload description file to pass",
    )
    parser.add_argument(
        "-t",
        "--targetList",
        dest="targetList",
        type=str,
        required=True,
        help="Input target list file to pass",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=False,
        default=None,
        help="Output file",
    )
    parser.add_argument(
        "-c",
        "--full_contrib",
        dest="contrib",
        default=False,
        required=False,
        help="produce full contribution output",
        action="store_true",
    )
    parser.add_argument(
        "-n",
        "--nThreads",
        dest="numberOfThreads",
        default=1,
        type=int,
        required=False,
        help="number of threads for parallel processing",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        default=False,
        required=False,
        help="log output on screen",
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--log",
        dest="log",
        default=False,
        required=False,
        help="log output on file",
        action="store_true",
    )
    parser.add_argument(
        "-P",
        "--plot",
        dest="plot",
        default=False,
        required=False,
        help="save target plots",
        action="store_true",
    )

    args = parser.parse_args()

    standard_pipeline(
        options=args.opt,
        target_list=args.targetList,
        output=args.output,
        plot=args.plot,
        full_contrib=args.contrib,
        debug=args.debug,
        n_thread=args.numberOfThreads,
        log=args.log,
    )



import logging
import unittest

from exorad.log import setLogLevel
from exorad.tasks.task import Task

setLogLevel(logging.DEBUG)


class TaskTestTask1(Task):
    class_name = 'testTask'

    def __init__(self):
        self.addTaskParam('input1', 'test input first parameter')
        self.addTaskParam('input2', 'test input second parameter')

    def execute(self):
        self.set_output(None)
        self.info('started')


class TaskTestTask2(Task):

    def __init__(self):
        pass

    def execute(self):
        pass


class TaskTest(unittest.TestCase):
    task1 = TaskTestTask1()
    task2 = TaskTestTask2()

    def test_input(self):
        with self.assertRaises(ValueError):
            self.task1(input1=1, input3=3)
        with self.assertRaises(TypeError):
            self.task1(1, 2)
        with self.assertRaises(ValueError):
            self.task1(input1=1, input2=2, input3=3)

        try:
            self.task1(input1=1, input2=2)
        except ValueError:
            self.fail("TaskTest() raised ValueErros unexpectedly!")



import logging
import os
import pathlib
import unittest

import astropy.units as u
from astropy.table import Column
from astropy.table import QTable
from test_options import payload_file

from exorad.log import setLogLevel
from exorad.output.hdf5 import HDF5Output
from exorad.tasks import LoadSource
from exorad.tasks.targetHandler import LoadTargetList
from exorad.tasks.targetHandler import PrepareTarget

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')

setLogLevel(logging.DEBUG)


class TargetTest(unittest.TestCase):
    target_list = os.path.join(data_dir, 'test_target.csv')

    def test_load_target_list(self):
        loadTargetList = LoadTargetList()
        targets = loadTargetList(target_list=self.target_list)
        with self.assertRaises(IOError):
            targets = loadTargetList(target_list='test_target.abc')

    def test_load_QTable_target_list(self):
        loadTargetList = LoadTargetList()
        names = Column(['test1', 'test2'], name='star name')
        masses = Column([1, 2] * u.M_sun, name='star M')
        temperatures = Column([5000, 6000] * u.K, name='star Teff')
        radii = Column([1, 2] * u.R_sun, name='star R')
        distances = Column([10, 12] * u.pc, name='star D')
        raw_targetlist = QTable(
            [names, masses, temperatures, radii, distances])

        targets = loadTargetList(target_list=raw_targetlist)

        distances = Column([10, 12], name='star D')
        raw_targetlist = QTable(
            [names, masses, temperatures, radii, distances])
        with self.assertRaises(TypeError):
            targets = loadTargetList(target_list=raw_targetlist)

    def test_write(self):
        loadTargetList = LoadTargetList()
        targets = loadTargetList(target_list=self.target_list)
        fname = 'test.h5'
        with HDF5Output(fname) as o:
            for target in targets.target:
                target.write(o)

        os.remove(fname)


class SourceTest(unittest.TestCase):
    loadTargetList = LoadTargetList()
    target_list = os.path.join(data_dir, 'test_target.csv')
    targets = loadTargetList(target_list=target_list)
    target = targets.target[0]
    loadSource = LoadSource()

    def test_planck(self):
        source = {'value': 'Planck'}
        target, sed = self.loadSource(target=self.target, source=source,
                                      wl_range=(0.45, 2.2) * u.um)
        target, sed = self.loadSource(target=self.target, source=source,
                                      wl_range=(0.45, 2.2))

    def test_phoenix(self):

        try:
            source = {'value': 'phoenix',
                      'StellarModels': {
                          'value': '/usr/local/project_data/sed/'}}
            target, sed = self.loadSource(target=self.target, source=source,
                                          wl_range=(0.45, 2.2) * u.um)
        except:
            print('indicate the stellar model directory to test this option')

    def test_custom(self):
        source = {'value': 'custom',
                  'CustomSed': {
                      'value': os.path.join(data_dir, 'customsed.csv')}}
        target, sed = self.loadSource(target=self.target, source=source,
                                      wl_range=(0.45, 2.2) * u.um)


class PrepareTargeTest(unittest.TestCase):
    from exorad.tasks import PreparePayload

    setLogLevel(logging.INFO)
    loadTargetList = LoadTargetList()
    target_list = os.path.join(data_dir, 'test_target.csv')
    targets = loadTargetList(target_list=target_list)
    target = targets.target[0]
    loadSource = LoadSource()
    target, sed = loadSource(target=target, source='planck',
                             wl_range=(0.45, 2.2))

    preparePayload = PreparePayload()
    payload, channels, (wl_min, wl_max) = preparePayload(
        payload_file=payload_file(),
        output=None)

    setLogLevel(logging.DEBUG)

    prepareTarget = PrepareTarget()
    target = prepareTarget(target=target, channels=channels)
    print(target.table)

    def test_write(self):
        print(self.target.to_dict())
        fname = 'test.h5'
        with HDF5Output(fname, append=True) as out:
            self.target.write(out)

        os.remove(fname)



"""
These parts of the code derives from TauREx3:
    @ARTICLE{2019arXiv191207759A,
           author = {{Al-Refaie}, Ahmed F. and {Changeat}, Quentin and {Waldmann}, Ingo P. and {Tinetti}, Giovanna},
            title = "{TauREx III: A fast, dynamic and extendable framework for retrievals}",
          journal = {arXiv e-prints},
         keywords = {Astrophysics - Instrumentation and Methods for Astrophysics, Astrophysics - Earth and Planetary Astrophysics},
             year = 2019,
            month = dec,
              eid = {arXiv:1912.07759},
            pages = {arXiv:1912.07759},
    archivePrefix = {arXiv},
           eprint = {1912.07759},
     primaryClass = {astro-ph.IM},
           adsurl = {https://ui.adsabs.harvard.edu/abs/2019arXiv191207759A},
          adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }

BSD 3-Clause License

Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the names of the copyright holders nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import datetime

import h5py
import hdfdict
from astropy.table import meta

from exorad import __author__
from exorad import __citation__
from exorad import __copyright__
from exorad import __license__
from exorad import __pkg_name__
from exorad import __title__
from exorad import __url__
from exorad import __version__
from exorad.output.hdf5.util import recursively_read_dict_contents
from exorad.output.output import Output
from exorad.output.output import OutputGroup

META_KEY = "__table_column_meta__"


class HDF5OutputGroup(OutputGroup):
    def __init__(self, entry):
        self.set_log_name()
        self._entry = entry

    def write_array(self, array_name, array, metadata=None):
        if isinstance(array, list):
            for idx, a in enumerate(array):
                self.write_array("{}{}".format(array_name, idx), a, metadata)
            return
        ds = self._entry.create_dataset(
            str(array_name), data=array, shape=array.shape, dtype=array.dtype
        )
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_table(self, table_name, table, metadata=None):
        table = _encode_mixins(table)
        if any(col.info.dtype.kind == "U" for col in table.itercols()):
            table = table.copy(copy_data=False)
            table.convert_unicode_to_bytestring()

        self._entry.create_dataset(str(table_name), data=table.as_array())
        header_yaml = meta.get_yaml_from_table(table)

        header_encoded = [h.encode("utf-8") for h in header_yaml]
        self._entry.create_dataset(
            str(table_name) + "." + META_KEY, data=header_encoded
        )

        tg = self._entry.create_group("{}_to_group".format(str(table_name)))
        for col in table.keys():
            tg_c = tg.create_group(str(col))
            tg_c.create_dataset("value", data=table[col])
            if table[col].unit != None:
                tg_c.create_dataset("unit", data=str(table[col].unit))

    def write_scalar(self, scalar_name, scalar, metadata=None):
        ds = self._entry.create_dataset(str(scalar_name), data=scalar)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_string(self, string_name, string, metadata=None):
        ds = self._entry.create_dataset(str(string_name), data=string)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def create_group(self, group_name):
        entry = None
        if self._entry:
            try:
                entry = self._entry.create_group(str(group_name))
            except ValueError:
                entry = self._entry[str(group_name)]
        return HDF5OutputGroup(entry)

    def write_string_array(self, string_name, string_array, metadata=None):
        asciiList = [n.encode("ascii", "ignore") for n in string_array]
        ds = self._entry.create_dataset(
            str(string_name), (len(asciiList), 1), "S64", asciiList
        )

        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def write_quantity(self, quantity_name, quantity):
        if quantity_name == "value":
            self._entry.create_dataset(
                "value",
                data=quantity.value,
            )
        else:
            qg_c = self._entry.create_group(str(quantity_name))
            qg_c.create_dataset(
                "value",
                data=quantity.value,
            )
            qg_c.create_dataset("unit", data=str(quantity.unit))
            pass


def _encode_mixins(tbl):
    from astropy.table import serialize
    from astropy.table.table import has_info_class
    from astropy import units as u
    from astropy.utils.data_info import MixinInfo, serialize_context_as

    try:
        import yaml
    except ImportError:
        for col in tbl.itercols():
            if (
                has_info_class(col, MixinInfo)
                and col.__class__ is not u.Quantity
            ):
                raise TypeError(
                    "cannot write type {} column '{}' "
                    "to HDF5 without PyYAML installed.".format(
                        col.__class__.__name__, col.info.name
                    )
                )

    with serialize_context_as("hdf5"):
        encode_tbl = serialize.represent_mixins_as_columns(tbl)

    return encode_tbl


class HDF5Output(Output):
    def __init__(self, filename, append=False):
        super().__init__()
        self.filename = filename
        self._append = append
        self.fd = None

    def open(self):
        self.fd = self._openFile(self.filename)

    def _openFile(self, fname):
        mode = "w"
        if self._append:
            mode = "a"

        attrs = {
            "file_name": fname,
            "file_time": datetime.datetime.now().isoformat(),
            "creator": self.__class__.__name__,
            "HDF5_Version": h5py.version.hdf5_version,
            "h5py_version": h5py.version.version,
            "program_name": str(__title__),
            "package name": str(__pkg_name__),
            "program_version": str(__version__),
            "author": str(__author__),
            "copyright": str(__copyright__),
            "license": str(__license__),
            "url": str(__url__),
            "citation": str(__citation__),
        }

        fd = h5py.File(fname, mode=mode)
        for key in attrs:
            fd.attrs[key] = attrs[key]

        if mode == "w" or "info" not in fd.keys():
            try:
                gd_ = fd["info"]
            except KeyError:
                gd_ = fd.create_group("info")
            gd = HDF5OutputGroup(gd_)
            gd.store_dictionary(attrs, "ExoRad")

        return fd

    def add_info(self, attrs, name=None):
        gd = self.create_group("info")
        gd.store_dictionary(attrs, name)

    def create_group(self, group_name):
        entry = None
        if self.fd:
            try:
                entry = self.fd.create_group(str(group_name))
            except ValueError:
                entry = self.fd[str(group_name)]
        return HDF5OutputGroup(entry)

    def close(self):
        if self.fd:
            self.fd.flush()
            self.fd.close()


def load(input_group):
    """
    Reads and convert an HDF5 group into a dictionary

    Parameters
    ----------
    input_group: HDF5Group
        input dumped group

    Returns
    -------
    dict
    """
    input_group = hdfdict.load(input_group, lazy=False)
    input_group = recursively_read_dict_contents(input_group)
    return input_group



import logging
import os
import pathlib
import unittest

from test_options import payload_file

import exorad.tasks as tasks
from exorad.log import setLogLevel

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')

setLogLevel(logging.DEBUG)


class PipelineTest(unittest.TestCase):
    loadOptions = tasks.LoadOptions()
    buildChannels = tasks.BuildChannels()
    buildInstrument = tasks.BuildInstrument()
    loadPayload = tasks.LoadPayload()
    loadTargetList = tasks.LoadTargetList()

    payload = loadOptions(filename=payload_file())
    wl_min, wl_max = payload['common']['wl_min']['value'], \
                     payload['common']['wl_max']['value']
    channels = buildChannels(payload=payload, write=False, output=None)
    targets = loadTargetList(
        target_list=os.path.join(data_dir, 'test_target.csv'))

    def test_full_pipe(self):
        prepareTarget = tasks.PrepareTarget()
        estimateForegrounds = tasks.EstimateForegrounds()
        propagateForegroundLight = tasks.PropagateForegroundLight()
        propagateTargetLight = tasks.PropagateTargetLight()
        estimateNoise = tasks.EstimateNoise()
        loadSource = tasks.LoadSource()

        for target in self.targets.target:
            target = prepareTarget(target=target, channels=self.channels)

            if 'foreground' in self.payload['common']:
                target = estimateForegrounds(
                    foregrounds=self.payload['common']['foreground'],
                    target=target,
                    wl_range=(self.wl_min, self.wl_max))
                target = propagateForegroundLight(channels=self.channels,
                                                  target=target)

            target, sed = loadSource(target=target,
                                     source=self.payload['common'][
                                         'sourceSpectrum'],
                                     wl_range=(self.wl_min, self.wl_max))
            target = propagateTargetLight(channels=self.channels,
                                          target=target)

            target = estimateNoise(target=target, channels=self.channels)

    def test_obsTarget(self):
        observeTarget = tasks.ObserveTarget()

        for target in self.targets.target:
            target = observeTarget(target=target, payload=self.payload,
                                   channels=self.channels,
                                   wl_range=(self.wl_min, self.wl_max))

    def test_obsTargetList(self):
        observeTargetList = tasks.ObserveTargetlist()

        targets = observeTargetList(targets=self.targets.target,
                                    payload=self.payload,
                                    channels=self.channels,
                                    wl_range=(self.wl_min, self.wl_max),
                                    plot=False, out_dir=None)



import astropy.units as u
import numpy as np
from astropy.table import meta
from astropy.table import QTable
from astropy.table import serialize
from astropy.table import Table

from exorad.models.signal import Sed
from exorad.models.signal import Signal

signal = {"W / (m2 um)": Sed}


def recursively_read_dict_contents(input_dict):
    """
    Will recursive read a dictionary, initializing quantities and table from a dictionary read from an hdf5 file.

    Parameters
    ----------
    input_dict : dict
        dictionary read from hdf5

    Returns
    --------
    dict
        Dictionary we want to use

    """
    new_keys = [k for k in input_dict.keys()]
    # if all(elem in new_keys for elem in ['wl_grid', 'data', 'time_grid']):
    #     wl = input['wl_grid']['value'] * u.Unit(input['wl_grid']['unit'])
    #     data = input['data']['value'] * u.Unit(input['data']['unit'])
    #     time = input['time_grid']['value'] * u.Unit(input['time_grid']['unit'])
    #     input = signal[str(input['data']['unit'])](wl, data, time)
    if all(elem in new_keys for elem in ["value", "unit"]):
        input_dict["value"] = input_dict["value"] * u.Unit(input_dict["unit"])

    if any(".__table_column_meta__" in elem for elem in new_keys):
        table_keys = [
            elem for elem in new_keys if ".__table_column_meta__" in elem
        ]
        table_keys = (elem.split(".")[0] for elem in table_keys)
        for k in table_keys:
            table = load_table(input_dict, k)
            input_dict[k] = table
    for key in new_keys:
        if isinstance(input_dict[key], dict):
            input_dict[key] = recursively_read_dict_contents(input_dict[key])
    return input_dict


def load_table(input_table, k):
    table = Table(np.array(input_table[k]))
    header = meta.get_header_from_yaml(
        h.decode("utf-8")
        for h in input_table["{}.__table_column_meta__".format(k)]
    )
    header_cols = {x["name"]: x for x in header["datatype"]}
    for col in table.columns.values():
        for attr in ("description", "format", "unit", "meta"):
            if attr in header_cols[col.name]:
                setattr(col, attr, header_cols[col.name][attr])
    table = serialize._construct_mixins_from_columns(table)
    try:
        header["meta"].pop("__serialized_columns__")
        table.meta = header["meta"]
    except KeyError:
        pass
    return table


def recursively_save_dict_contents_to_output(output, dic):
    """
    Will recursive write a dictionary into output.

    Parameters
    ----------
    output :
        Group (or root) in output file to write to

    dic : :obj:`dict`
        Dictionary we want to write

    """

    for key, item in dic.items():
        try:
            store_thing(output, key, item)
        except TypeError:
            raise ValueError("Cannot write %s type" % type(item))
    return


def store_thing(output, key, item):
    if isinstance(item, u.Quantity):
        output.write_quantity(key, item)
    elif isinstance(
        item,
        (
            float,
            int,
            np.int64,
            np.int32,
            np.float64,
        ),
    ):
        output.write_scalar(key, item)
    elif isinstance(item, np.ndarray):
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            output.write_array(key, item)
    elif isinstance(item, (str,)):
        output.write_string(key, item)
    elif isinstance(item, (Table, QTable)):
        output.write_table(key, item)
    elif isinstance(item, Signal):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item.to_dict())
    elif isinstance(
        item,
        (
            list,
            tuple,
        ),
    ):
        if isinstance(item, tuple):
            item = list(item)
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            try:
                output.write_array(key, np.array(item))

            except TypeError:
                for idx, val in enumerate(item):
                    new_key = "{}{}".format(key, idx)
                    store_thing(output, new_key, val)

    elif isinstance(item, dict):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item)
    else:
        raise TypeError
    return



import copy
from abc import abstractmethod

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import QTable
from scipy.interpolate import interp1d

from exorad.log.logger import Logger
from exorad.models.optics.opticalPath import OpticalPath
from exorad.models.signal import Signal
from exorad.models.utils import get_wl_col_name
from exorad.utils.diffuse_light_propagation import convolve_with_slit
from exorad.utils.diffuse_light_propagation import integrate_light
from exorad.utils.diffuse_light_propagation import prepare
from exorad.utils.passVal import PassVal


class Instrument(Logger):
    """
    Abstract Instrument class. Contains the basic function that every instrument needs.

    Parameters
    -----------
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None

    Attributes
    ----------
    name: str
        instrument name
    description: dict
        instrument description dictionary
    payload: dict
        main payload. Default is None
    table: QTable
        contain the output grid for the instrument
    built_instr: dict
        contains the instrument parameters needed to propagate the signal

    Raises
    -------
    ValueError:
        if you try to build a loaded payload.
    """

    def __init__(self, name, description, payload=None):
        self.set_log_name()
        self.name = name
        self.description = description
        self.payload = payload
        self.table = QTable()
        self.built_instr = {}
        self.debug("{} initialized".format(self.name))
        self.loaded = False
        self.opticalPath = None

    def load(self, table, built_instr):
        """
        it loads the instrument parameters already processed from a file
        """
        self.table = table
        self.built_instr = built_instr
        self.loaded = True
        self.info("{} loaded".format(self.name))

    def write(self, output):
        """
        it writes the instrument parameters already processed from a file
        """
        inst_out = output.create_group(self.name)
        inst_out.write_table(
            self.name,
            self.table,
        )
        inst_out.store_dictionary(self.description, group_name="description")
        inst_out.store_dictionary(self.built_instr, group_name="built_instr")
        self.info("instrument saved")
        return output

    def build(self):
        """
        check if it can build the instrument and populate the output table and dict.
        Then run the builder
        """

        if self.loaded:
            self.error("You cannot build a loaded instrument")
            # to build the instrument we need to be careful about the elements order, especially for optic emission!
            # you cannot load the description and run it again because it's not an ordered dictionary anymore
            raise ValueError("You cannot build a loaded instrument")
        else:
            self.builder()
            self.build_optical_path()

    def build_optical_path(self):
        """
        it builds the instrument optical path
        """

        self.info("building optical path")
        # what wl do I wanna use here?
        wl_grid = (
            np.logspace(
                np.log10(
                    self.description["detector"]["wl_min"]["value"].value
                ),
                np.log10(
                    self.description["detector"]["cut_off"]["value"].value
                ),
                PassVal.working_R,
            )
            * u.um
        )

        common_optical_path = OpticalPath(wl=wl_grid, description=self.payload)
        channel_optical_path = OpticalPath(
            wl=wl_grid, description=self.description
        )
        channel_optical_path.prepend_optical_elements(
            common_optical_path.optical_element_dict
        )
        self.opticalPath = channel_optical_path

        transmission_table = self.opticalPath.build_transmission_table()
        optical_path_dict = {"transmission_table": transmission_table}
        # We need to rebin TR to the final one
        self.table["TR"], transmission_data = self._get_transmission(
            wl_grid, transmission_table["total"]
        )
        self._add_data_to_built(
            "transmission_data", transmission_data.to_dict()
        )

        self.opticalPath.chain()
        if self.opticalPath.slit_width is not None:
            self._add_data_to_built(
                "slit_width", self.opticalPath.slit_width.to(u.um)
            )

        self.table = self.opticalPath.compute_signal(
            self.table, self.built_instr
        )
        optical_path_dict[
            "max_signal_per_pixel"
        ] = self.opticalPath.max_signal_per_pixel
        optical_path_dict["signal_table"] = self.opticalPath.signal_table
        optical_path_dict["radiance_table"] = self.opticalPath.radiance_table
        self._add_data_to_built("optical_path", optical_path_dict)

    @abstractmethod
    def builder(self):
        """
        build the instrument and populate the output table and dict
        """
        pass

    @abstractmethod
    def propagate_target(self, target):
        """
        propagates the target light through the instrument
        """
        pass

    def propagate_diffuse_foreground(self, target):
        """
        propagate diffuse foreground sources, starting from zodiacal background
        """
        import copy

        self.debug("diffuse bkg propagation")
        out = QTable()
        (
            total_max_signal,
            total_signal,
            wl_table,
            A,
            qe,
            omega_pix,
            transmission,
        ) = prepare(self.table, self.built_instr, self.description)
        foregrounds = list(target.foreground.keys())
        foregrounds = reversed(foregrounds)
        for i, frg in enumerate(foregrounds):
            self.debug("propagating {}".format(frg))
            radiance = copy.deepcopy(target.foreground[frg])
            self.debug("{} radiance . {}".format(frg, radiance.data))

            if hasattr(frg, "transmission"):
                frg.transmission.spectral_rebin(transmission.wl_grid)
                transmission.data *= frg.transmission.data
                self.debug("added {} transmission".format(frg))

            transmission.spectral_rebin(radiance.wl_grid)
            radiance.data *= transmission.data

            # for other_el in foregrounds[i+1:]:
            #     if hasattr(target.foreground[other_el], 'transmission'):
            #         self.debug('passing through {}'.format(other_el))
            #         radiance.data *= target.foreground[other_el].transmission
            if "slit_width" in self.built_instr:
                max_signal_per_pix, signal = convolve_with_slit(
                    self.description,
                    self.built_instr,
                    A,
                    self.table,
                    omega_pix,
                    qe,
                    radiance,
                )
            else:
                qe.spectral_rebin(radiance.wl_grid)
                radiance.data *= (
                    omega_pix
                    * A
                    * qe.data
                    * (qe.wl_grid / const.c / const.h).to(1.0 / u.W / u.s)
                    * u.count
                )
                # try:
                #     signal = np.trapz(radiance.data * self._window_function(radiance), x=radiance.wl_grid).to(u.ct / u.s)
                #     signal *= self.built_instr['window_spatial_width']
                # except KeyError:
                max_signal_per_pix, signal = integrate_light(
                    radiance, radiance.wl_grid, self.built_instr
                )
            total_signal = copy.deepcopy(signal)
            self.debug("sed : {}".format(total_signal))
            total_max_signal = copy.deepcopy(max_signal_per_pix)

            out["{}_signal".format(frg)] = total_signal
            out["{}_MaxSignal_inPixel".format(frg)] = total_max_signal
        return out

    def _bin_signal(self, wl, signal, leftbin, rightbin):
        bsig = [
            np.mean(signal[np.logical_and(wl >= wlow, wl < whigh)])
            for wlow, whigh in zip(leftbin, rightbin)
        ]
        return u.Quantity(bsig)

    def _get_transmission(self, wl_grid, tr_data=None):
        tr_func = interp1d(
            wl_grid,
            tr_data,
            assume_sorted=False,
            fill_value=0.0,
            bounds_error=False,
        )

        transmission_data = Signal(wl_grid, tr_data)
        # self.debug('transmission channel data : {} {}'.format(transmission_data.wl_grid, transmission_data.data))

        transmission = self._bin_signal(
            self.table["Wavelength"],
            tr_func(self.table["Wavelength"]),
            self.table["LeftBinEdge"],
            self.table["RightBinEdge"],
        )

        if self.payload["optics"]["ForceChannelWlEdge"]["value"]:
            self.debug("force channel wl edge enabled")
            idx = np.logical_or(
                self.table["Wavelength"]
                < self.description["wl_min"]["value"].to(
                    self.table["Wavelength"].unit
                ),
                self.table["Wavelength"]
                > self.description["wl_max"]["value"].to(
                    self.table["Wavelength"].unit
                ),
            )
            transmission[idx] = 0.0
            idx = np.logical_or(
                transmission_data.wl_grid
                < self.description["wl_min"]["value"].to(
                    self.table["Wavelength"].unit
                ),
                transmission_data.wl_grid
                > self.description["wl_max"]["value"].to(
                    self.table["Wavelength"].unit
                ),
            )
            transmission_data.data[idx] = 0.0

        self.debug("transmission in channel : {}".format(transmission))
        return transmission, transmission_data

    def _get_qe(self):
        if "datafile" in self.description["detector"]["qe"]:
            self.debug("QE data file found")
            wl_colname = get_wl_col_name(
                self.description["detector"]["qe"]["data"]
            )
            qe_data = Signal(
                self.description["detector"]["qe"]["data"][wl_colname],
                self.description["detector"]["qe"]["data"][self.name],
            )
        else:
            wl_grid = (
                np.logspace(
                    np.log10(
                        self.description["detector"]["wl_min"]["value"]
                        .to(u.um)
                        .value
                    ),
                    np.log10(
                        self.description["detector"]["cut_off"]["value"]
                        .to(u.um)
                        .value
                    ),
                    PassVal.working_R,
                )
                * u.um
            )
            qe_data = Signal(
                wl_grid,
                self.description["detector"]["qe"]["value"]
                * np.ones(wl_grid.size),
            )

        qe_func = interp1d(
            qe_data.wl_grid,
            qe_data.data,
            assume_sorted=False,
            fill_value=0.0,
            bounds_error=False,
        )
        qe = self._bin_signal(
            self.table["Wavelength"],
            qe_func(self.table["Wavelength"]),
            self.table["LeftBinEdge"],
            self.table["RightBinEdge"],
        )
        self.debug("qe in channel : {}".format(qe))
        return qe, qe_data

    def _get_efficiency(self, wl, target):
        # qe = interp1d(wl, self.built_instr['qe_data']['wl_grid'],
        #               self.built_instr['qe_data']['data'], left=0.0, right=0.0)
        # transmission = interp1d(wl, self.built_instr['transmission_data']['wl_grid'],
        #                         self.built_instr['transmission_data']['data'], left=0.0, right=0.0)
        # wave_window = np.ones(self.table['Wavelength'])
        qe = Signal(
            self.built_instr["qe_data"]["wl_grid"]["value"]
            * u.Unit(self.built_instr["qe_data"]["wl_grid"]["unit"]),
            self.built_instr["qe_data"]["data"]["value"],
        )
        qe.spectral_rebin(wl)

        transmission = Signal(
            self.built_instr["transmission_data"]["wl_grid"]["value"]
            * u.Unit(self.built_instr["transmission_data"]["wl_grid"]["unit"]),
            self.built_instr["transmission_data"]["data"]["value"],
        )
        transmission.spectral_rebin(wl)

        if hasattr(target, "skyTransmission"):
            target_transmission = copy.deepcopy(target.skyTransmission)
            self.table["sky TR"], _ = self._get_transmission(
                target_transmission.wl_grid, target_transmission.data
            )
            target_transmission.spectral_rebin(wl)
            transmission.data *= target_transmission.data

        wave_window = np.ones(wl.size)
        return qe.data, transmission.data, wave_window

    def _add_data_to_built(self, name, data):
        self.built_instr[name] = data

    def _window_function(self, signal):
        window_function = []
        for wld, wlu in zip(
            self.table["LeftBinEdge"], self.table["RightBinEdge"]
        ):
            mask = np.logical_and(
                signal.wl_grid >= wld, signal.wl_grid < wlu
            ).astype(float)
            window_function.append(mask)
        window_function = np.array(window_function)
        self.debug("window function: {}".format(window_function))
        return window_function



import logging
import os
import pathlib
import unittest

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from conf import skip_plot
from test_options import payload_file

from exorad.log import setLogLevel
from exorad.models.instruments import Photometer
from exorad.models.instruments import Spectrometer
from exorad.models.optics.opticalPath import OpticalPath
from exorad.tasks.loadOptions import LoadOptions

path = pathlib.Path(__file__).parent.absolute()
data_dir = os.path.join(path.parent.absolute(), 'examples')

loadOptions = LoadOptions()
options = loadOptions(filename=payload_file())


class InstrumentDiffuseLightTest(unittest.TestCase):

    def test_transmission(self):

        wl_grid = np.logspace(np.log10(
            options['channel']['Phot']['detector']['wl_min']['value'].value),
                              np.log10(options['channel']['Phot']['detector'][
                                           'cut_off']['value'].value),
                              6000) * u.um
        telescope = OpticalPath(wl=wl_grid, description=options)
        tel = telescope.chain()
        phot = OpticalPath(wl=wl_grid, description=options['channel']['Phot'])
        phot.prepend_optical_elements(telescope.optical_element_dict)
        phot.build_transmission_table()
        phot = phot.chain()
        #
        wl_grid = np.logspace(np.log10(
            options['channel']['Spec']['detector']['wl_min']['value'].value),
                              np.log10(options['channel']['Spec']['detector'][
                                           'cut_off']['value'].value),
                              6000) * u.um
        telescope = OpticalPath(wl=wl_grid, description=options)
        tel = telescope.chain()
        spec = OpticalPath(wl=wl_grid, description=options['channel']['Spec'])
        spec.prepend_optical_elements(telescope.optical_element_dict)
        spec.build_transmission_table()

        if not skip_plot:
            fig = plt.figure(11)
            tab = spec.transmission_table
            for o in tab.keys():
                if o == 'Wavelength': continue
                plt.plot(tab['Wavelength'], tab[o], label=o)
            plt.title('test_optical.test_transmission')
            plt.legend()
            plt.show()

        spec = spec.chain()

        if not skip_plot:
            fig, ax = plt.subplots(1, 1)
            for o in phot:
                phot[o].plot(fig=fig, ax=ax, label=o, yscale='log')
            for o in spec:
                spec[o].plot(fig=fig, ax=ax, label=o, yscale='log')
            plt.legend()
            plt.show()

    def test_signal(self):
        setLogLevel(logging.INFO)

        photometer = Photometer('Phot', options['channel']['Phot'], options)
        photometer.build()

        spectrometer = Spectrometer('Spec', options['channel']['Spec'],
                                    options)
        spectrometer.build()

        setLogLevel(logging.DEBUG)

        wl_grid = np.logspace(np.log10(
            options['channel']['Phot']['detector']['wl_min']['value'].value),
                              np.log10(options['channel']['Phot']['detector'][
                                           'cut_off']['value'].value),
                              6000) * u.um
        telescope = OpticalPath(wl=wl_grid, description=options)
        telescope.chain()
        phot = OpticalPath(wl=wl_grid, description=options['channel']['Phot'])
        phot.prepend_optical_elements(telescope.optical_element_dict)
        phot.build_transmission_table()
        phot.chain()
        phot.compute_signal(photometer.table, photometer.built_instr)
        print(phot.signal_table)

        wl_grid = np.logspace(np.log10(
            options['channel']['Spec']['detector']['wl_min']['value'].value),
                              np.log10(options['channel']['Spec']['detector'][
                                           'cut_off']['value'].value),
                              6000) * u.um
        telescope = OpticalPath(wl=wl_grid, description=options)
        telescope.chain()
        spec = OpticalPath(wl=wl_grid, description=options['channel']['Spec'])
        spec.prepend_optical_elements(telescope.optical_element_dict)
        spec.build_transmission_table()
        spec.chain()
        spec.compute_signal(spectrometer.table, spectrometer.built_instr)
        print(spec.signal_table)

        if not skip_plot:
            fig, ax = plt.subplots(1, 1)
            wl = spectrometer.table['Wavelength']
            for col in spec.signal_table.keys():
                ax.plot(wl, spec.signal_table[col], alpha=0.5, label=col)
            ax.set_yscale('log')
            ax.set_title('test_optical.test_signal')
            plt.legend()
            plt.show()

    def test_values(self):

        # set constant qe
        options['channel']['Spec']['detector']['qe'] = {'value': 0.7}

        # prepare channel
        setLogLevel(logging.INFO)
        wl_grid = np.logspace(np.log10(
            options['channel']['Spec']['detector']['wl_min']['value'].value),
                              np.log10(options['channel']['Spec']['detector'][
                                           'cut_off']['value'].value),
                              6000) * u.um
        spectrometer = Spectrometer('Spec', options['channel']['Spec'],
                                    options)
        spectrometer.build()
        spec = OpticalPath(wl=wl_grid, description=options['channel']['Spec'])
        spec.build_transmission_table()
        spec.chain()
        spec.compute_signal(spectrometer.table, spectrometer.built_instr)

        setLogLevel(logging.DEBUG)

        # prepare elements to compute
        D1 = options['channel']['Spec']['optics']['opticalElement']['D1']
        M5 = options['channel']['Spec']['optics']['opticalElement']['M5']
        M6 = options['channel']['Spec']['optics']['opticalElement']['M6']
        optics = options['channel']['Spec']['optics']['opticalElement'][
            'optics']
        detector = options['channel']['Spec']['optics']['opticalElement'][
            'detector']
        qe = 0.7
        D = (options['channel']['Spec']['detector']['delta_pix']['value'])
        Windows = spectrometer.built_instr['window_size_px']
        Windows_spatial = spectrometer.built_instr['window_spatial_width']

        from exorad.utils.exolib import planck
        import astropy.constants as const
        from astropy.table import Table
        check_table = Table()

        # check detector: pi acceptance angle
        detector_radiance = planck(wl=wl_grid,
                                   T=detector['temperature']['value']).to(
            u.W / (u.m ** 2 * u.micron * u.sr))
        detector_radiance *= D ** 2 * np.pi * u.sr * qe * (
                    wl_grid / const.c / const.h).to(
            1. / u.W / u.s) * u.count
        detector_signal = (np.trapz(detector_radiance, x=wl_grid)).to(
            u.count / u.s)
        check_table['detector signal'] = detector_signal * Windows

        # check optic box: pi-omega_pix acceptance angle
        from exorad.utils.exolib import OmegaPix
        omega_pix = OmegaPix(
            options['channel']['Spec']['Fnum_x']['value'].value,
            options['channel']['Spec']['Fnum_y']['value'].value)
        optics_radiance = planck(wl=wl_grid,
                                 T=optics['temperature']['value']).to(
            u.W / (u.m ** 2 * u.micron * u.sr))
        optics_radiance *= D ** 2 * (np.pi * u.sr - omega_pix) * qe * (
                    wl_grid / const.c / const.h).to(
            1. / u.W / u.s) * u.count
        optics_signal = (np.trapz(optics_radiance, x=wl_grid)).to(
            u.count / u.s)
        check_table['optics signal'] = optics_signal * Windows

        # check M6: omega_pix acceptance angle
        M6_radiance = planck(wl=wl_grid, T=M6['temperature']['value']).to(
            u.W / (u.m ** 2 * u.micron * u.sr))
        M6_radiance *= M6['emissivity']['value']
        M6_radiance *= D ** 2 * omega_pix * qe * (
                    wl_grid / const.c / const.h).to(
            1. / u.W / u.s) * u.count
        M6_signal = (np.trapz(M6_radiance, x=wl_grid)).to(
            u.count / u.s)
        check_table['M6 signal'] = M6_signal * Windows

        # check M5: omega_pix acceptance angle, transmission applied (also D1) and slit
        slit_width = spectrometer.built_instr['slit_width']
        wl_pix = spectrometer.built_instr['wl_pix_center']
        dwl_pic = spectrometer.built_instr['pixel_bandwidth']
        ch_table = spectrometer.table
        slit_kernel = np.ones(int(slit_width / D.to(u.um)))

        M5_radiance = planck(wl=wl_grid, T=M5['temperature']['value']).to(
            u.W / (u.m ** 2 * u.micron * u.sr))
        M5_radiance *= M5['emissivity']['value'] * M6['transmission'][
            'value'] * D1['transmission']['value']

        wl_min, wl_max = D1['wl_min']['value'], D1['wl_max']['value']
        idx = np.where(wl_grid < wl_min)
        M5_radiance[idx] = 0.0
        idx = np.where(wl_grid > wl_max)
        M5_radiance[idx] = 0.0
        M5_radiance *= D ** 2 * omega_pix * qe * (
                    wl_grid / const.c / const.h).to(
            1. / u.W / u.s) * u.count
        from exorad.utils.exolib import rebin
        _, M5_radiance = rebin(wl_pix, wl_grid, M5_radiance)
        M5_signal_tmp = (
            np.convolve(M5_radiance * dwl_pic, slit_kernel, 'same')).to(
            u.count / u.s)
        idx = [np.logical_and(wl_pix > wlow, wl_pix <= whigh)
               for wlow, whigh in
               zip(ch_table['LeftBinEdge'], ch_table['RightBinEdge'])]
        M5_signal = [M5_signal_tmp[idx[k]].value.sum() for k in
                     range(len(idx))]
        check_table['M5 signal'] = M5_signal * u.ct / u.s * Windows_spatial

        spec.signal_table.remove_column('D1 signal')

        if not skip_plot:
            fig, ax = plt.subplots(1, 1)
            wl = spectrometer.table['Wavelength']
            for col in spec.signal_table.keys():
                ax.plot(wl, spec.signal_table[col], alpha=0.5, lw=1, label=col)
                ax.plot(wl, check_table[col], alpha=0.5, ls=':', lw=3,
                        label='{} check'.format(col))
            ax.set_yscale('log')
            fig.suptitle('test_optical.test_values 1')
            ax.legend()
            plt.show()

            fig, ax = plt.subplots(1, 1)
            wl = spectrometer.table['Wavelength']
            for col in spec.signal_table.keys():
                ax.plot(wl, spec.signal_table[col] / check_table[col],
                        alpha=0.5, label=col)
            ax.set_yscale('log')
            fig.suptitle('test_optical.test_values 2')
            ax.legend()
            plt.show()

        for col in spec.signal_table.keys():
            # check absolute equality up to 1e-40
            np.testing.assert_almost_equal(spec.signal_table[col].value,
                                           check_table[col], decimal=40,
                                           err_msg='',
                                           verbose=True)

        for col in spec.signal_table.keys():
            # check relative equality up to 1%
            np.testing.assert_almost_equal(
                spec.signal_table[col].value / check_table[col],
                np.ones(len(check_table[col])), decimal=2, err_msg='',
                verbose=True)



import importlib.metadata as metadata
import os.path
from datetime import date

# load package info
__pkg_name__ = metadata.metadata("exorad")["Name"]
__url__ = metadata.metadata("exorad")["Project-URL"]
__author__ = metadata.metadata("exorad")["Author"]
__email__ = metadata.metadata("exorad")["Author_email"]
__license__ = metadata.metadata("exorad")["license"]
__summary__ = metadata.metadata("exorad")["Summary"]

# load package commit number
try:
    __base_dir__ = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
except NameError:
    __base_dir__ = None

__commit__ = None
__branch__ = None
if __base_dir__ is not None and os.path.exists(
    os.path.join(__base_dir__, ".git")
):
    git_folder = os.path.join(__base_dir__, ".git")
    with open(os.path.join(git_folder, "HEAD")) as fp:
        ref = fp.read().strip()
    ref_dir = ref[5:]
    __branch__ = ref[16:]
    try:
        with open(os.path.join(git_folder, ref_dir)) as fp:
            __commit__ = fp.read().strip()
    except FileNotFoundError:
        __commit__ = None

__title__ = "ExoRad2"

__copyright__ = "2020-{:d}, {}".format(date.today().year, __author__)
__citation__ = "Mugnai et al., 2020, 'ArielRad: the ARIEL radiometric model', Exp. Astron, 50, 303-328"

from .exorad import standard_pipeline


from exorad.utils.version_control import VersionControl

VersionControl()



from .targetHandler import UpdateTargetTable
from .task import Task


class PropagateTargetLight(Task):
    """
    Given the target and a channel dict, it propagates the target light over the channels

    Parameters
    ----------
    channels: dict
        main payload. Default is None
    target: Target

    Returns
    -------
    QTable:
        return the output table of propagated signal

    Examples
    --------

    """

    def __init__(self):
        self.addTaskParam("channels", "payload channels dict")
        self.addTaskParam("target", "Target class to observe")

    def execute(self):
        self.info("propagating target light")
        from exorad.utils.util import vstack_tables

        target = self.get_task_param("target")
        channels = self.get_task_param("channels")
        self.debug("detectors found : {}".format(channels.keys()))
        table_list = []
        for ch in self.get_task_param("channels"):
            self.debug("propagating target in {}".format(ch))
            table_list.append(channels[ch].propagate_target(target))
        table = vstack_tables(table_list)
        updateTargetTable = UpdateTargetTable()
        target = updateTargetTable(target=target, table=table)
        self.set_output(target)


class PropagateForegroundLight(Task):
    """
    Given the target and a channel dict, it propagates the target foreground light over the channels

    Parameters
    ----------
    channels: dict
        main payload. Default is None
    target: Target

    Returns
    -------
    QTable:
        return the output table of propagated signal

    Examples
    --------

    """

    def __init__(self):
        self.addTaskParam("channels", "payload channels dict")
        self.addTaskParam("target", "Target class to observe")

    def execute(self):
        self.info("propagating target foreground light")
        from exorad.utils.util import vstack_tables

        target = self.get_task_param("target")
        channels = self.get_task_param("channels")
        self.debug("detectors found : {}".format(channels.keys()))
        table_list = []
        for ch in self.get_task_param("channels"):
            self.debug("propagating target background in {}".format(ch))
            table_list.append(
                channels[ch].propagate_diffuse_foreground(target)
            )
        table = vstack_tables(table_list)
        updateTargetTable = UpdateTargetTable()
        target = updateTargetTable(target=target, table=table)
        self.set_output(target)



import os

import astropy.constants as const
import astropy.units as u
import numpy as np
import pandas as pd
from astropy.table import QTable
from scipy.interpolate import interp1d

from .instrument import Instrument
from exorad.models.signal import CountsPerSeconds
from exorad.models.signal import CustomSignal
from exorad.models.signal import Signal
from exorad.utils.exolib import binnedPSF
from exorad.utils.exolib import find_aperture_radius
from exorad.utils.exolib import paosPSF
from exorad.utils.exolib import pixel_based_psf


class Spectrometer(Instrument):
    """
    Spectrometer instrument class.
    """

    def _wavelength_table(self):
        # check if R is defined
        if "targetR" not in self.description:
            self.warning("Channel targetR missing: native R is assumed")
            self.description["targetR"] = {"value": "native"}

        # define the wavelength grid
        if "value" in self.description["targetR"].keys():
            self.debug(
                "spectral resolution set to {}".format(
                    self.description["targetR"]["value"]
                )
            )
            # native R
            if self.description["targetR"]["value"] == "native":
                wl_bin_c = self.built_instr["wl_pix_center"]
                wl_bin_width = self.built_instr["pixel_bandwidth"]
                idx = np.argsort(wl_bin_c)
                wl_bin_c = wl_bin_c[idx]
                wl_bin_width = wl_bin_width[idx]
                wl_bin = wl_bin_c - 0.5 * wl_bin_width
                wl_bin = np.append(
                    wl_bin, wl_bin_c[-1] + 0.5 * wl_bin_width[-1]
                )
            # fixed R
            else:
                number_of_spectral_bins = (
                    np.ceil(
                        np.log(
                            self.description["wl_max"]["value"]
                            / self.description["wl_min"]["value"]
                        )
                        / np.log(
                            1.0 + 1.0 / self.description["targetR"]["value"]
                        )
                    )
                    + 1
                )
                wl_bin = self.description["wl_min"]["value"] * (
                    1.0 + 1.0 / self.description["targetR"]["value"]
                ) ** np.arange(number_of_spectral_bins)
                wl_bin_c = 0.5 * (wl_bin[0:-1] + wl_bin[1:])
                wl_bin_width = wl_bin[1:] - wl_bin[0:-1]
        # wavelength dependent R
        elif "data" in self.description["targetR"].keys():
            res_data = pd.read_csv(
                self.description["targetR"]["data"]["value"], sep="\t"
            )
            get_res = interp1d(res_data["W"], res_data["R"])
            w1 = self.description["wl_min"]["value"].value
            w2 = w1
            wl_bin = np.array([w1])
            while w2 < self.description["wl_max"]["value"].value:
                w1 = w2
                res = get_res(w1)
                w2 = np.round((w1 + w1 / res), 4)
                wl_bin = np.append(wl_bin, w2)
            if (
                self.description["wl_max"]["value"].value - wl_bin[-2]
                < (wl_bin[-2] - wl_bin[-3]) / 2.0
            ):  # remove small bin at end of channel
                wl_bin[-2] = self.description["wl_max"]["value"].value
                wl_bin = wl_bin[:-1]
            wl_bin_c = (
                (np.array(wl_bin[:-1]) + np.array(wl_bin[1:])) / 2 * u.micron
            )
            wl_bin_width = (
                np.array(wl_bin[1:]) - np.array(wl_bin[:-1])
            ) * u.micron
        else:
            self.error("Channel targetR format unsupported.")
            raise KeyError("Channel targetR format unsupported.")

        # preparing the table
        self.table["chName"] = [self.name] * wl_bin_c.size
        self.table["Wavelength"] = wl_bin_c
        self.table["Bandwidth"] = wl_bin_width
        self.table["LeftBinEdge"] = (
            self.table["Wavelength"] - 0.5 * self.table["Bandwidth"]
        )
        self.table["RightBinEdge"] = (
            self.table["Wavelength"] + 0.5 * self.table["Bandwidth"]
        )
        self.debug("wavelength table: \n{}".format(self.table))
        return wl_bin

    def builder(self):
        self.info("building {}".format(self.name))

        # building pixel grid
        try:
            wl_solution_data = CustomSignal(
                self.description["wlSolution"]["data"]["Wavelength"],
                self.description["wlSolution"]["data"]["x"],
                self.description["wlSolution"]["data"]["x"].unit,
            )
        except KeyError:
            self.error("Wavelength solution not indicated")
            raise

        self._add_data_to_built("wl_solution_data", wl_solution_data.to_dict())
        wl_sol_func = interp1d(
            wl_solution_data.data,
            wl_solution_data.wl_grid,
            assume_sorted=False,
            fill_value="extrapolate",
            bounds_error=False,
        )

        wl_sol_func_reverse = interp1d(
            wl_solution_data.wl_grid,
            wl_solution_data.data,
            assume_sorted=False,
            fill_value="extrapolate",
            bounds_error=False,
        )

        first_pixel = (
            wl_sol_func_reverse(self.description["wl_min"]["value"])
            * wl_solution_data.data.unit
        )
        last_pixel = (
            wl_sol_func_reverse(self.description["wl_max"]["value"])
            * wl_solution_data.data.unit
        )
        self.debug(
            "first pixel: {}, last pixel: {}".format(first_pixel, last_pixel)
        )

        if last_pixel < first_pixel:
            last_pixel, first_pixel = first_pixel, last_pixel

        delta = self.description["detector"]["delta_pix"]["value"].to(
            first_pixel.unit
        )

        # coordinates of detector  pixel centres
        coord_pix_center = (
            np.arange(first_pixel.value, last_pixel.value, delta.value)
            * delta.unit
        )
        coord_pix_center = np.flip(coord_pix_center)
        self.debug("pix_center: {}".format(coord_pix_center))
        self._add_data_to_built("pix_center", coord_pix_center)
        # wavelength sampled by the pixels
        pixel_wavelength = (
            wl_sol_func(coord_pix_center) * wl_solution_data.wl_grid.unit
        )
        self.debug("wl_pix_center: {}".format(pixel_wavelength))
        self._add_data_to_built("wl_pix_center", pixel_wavelength)
        pixel_bandwidth = (
            np.abs(
                wl_sol_func(coord_pix_center - 0.5 * delta)
                - wl_sol_func(coord_pix_center + 0.5 * delta)
            )
            * delta.unit
        )
        self._add_data_to_built("pixel_bandwidth", pixel_bandwidth)

        # initializing channel table
        wl_bin = self._wavelength_table()
        self._add_data_to_built("wl_bin", wl_bin)
        self.table["QE"], qe_data = self._get_qe()
        self._add_data_to_built("qe_data", qe_data.to_dict())

        # PSF gain
        _gain_prf = np.empty(10, dtype=float)
        wl_prf = np.linspace(
            pixel_wavelength.min(), pixel_wavelength.max(), _gain_prf.size
        )
        prfs = []
        for k, wl in enumerate(wl_prf):
            psf_file = None
            psf_format = None
            if "PSF" in self.description.keys():
                psf_file = self.description["PSF"]["value"]
                if "format" in self.description["PSF"].keys():
                    psf_format = self.description["PSF"]["format"]["value"]

            if psf_format == "pixel_based":
                # If psf is pixel based it simply loads the image
                prf, pixel_prf, extent = pixel_based_psf(wl, delta, psf_file)

            elif psf_format == "paos":
                prf, pixel_prf, extent = paosPSF(
                    wl=[wl] * u.micron, delta_pix=delta, filename=psf_file
                )
            else:
                prf, pixel_prf, extent = binnedPSF(
                    self.description["Fnum_x"]["value"],
                    self.description["Fnum_y"]["value"],
                    [wl] * u.micron,
                    delta,
                    psf_file,
                )

            _gain_prf[k] = np.max(
                prf.sum(axis=0) / pixel_prf.sum(axis=0).max()
            )
            prfs += [prf]

            if "WFErms" in self.description.keys():
                self.debug("WFE found")
                Strehl = np.exp(
                    -(
                        (2 * np.pi * self.description["WFErms"]["value"] / wl)
                        ** 2
                    )
                )
            else:
                Strehl = 1.0
            _gain_prf[k] *= Strehl
        self.debug(_gain_prf)
        gain_prf_data = Signal(wl_prf, _gain_prf)
        self._add_data_to_built("gain_prf_data", gain_prf_data.to_dict())

        # window sizes
        pixel_window_edge = (
            wl_sol_func_reverse(wl_bin) * wl_solution_data.data.unit
        )
        window_spectral_width = (
            np.abs(pixel_window_edge[1:] - pixel_window_edge[0:-1]) / delta
        ).to("")

        if "EncESolution" in self.description.keys():
            # if encircled energy radii are provided, load from file
            # todo define spatial apertures
            self.debug("Encircle energy solution found")
            EncE_sol = interp1d(
                self.description["EncESolution"]["data"]["Wavelength"],
                self.description["EncESolution"]["data"]["EE"],
                assume_sorted=False,
                fill_value="extrapolate",
            )
            radius = EncE_sol(self.table["Wavelength"])
        # elif os.path.isdir(self.description['PSF']['value']) and (
        #         'EnE' in self.description.keys()):
        #     # if a list of psf and the desired encircled are provided, estimate the radii
        #     enc = []
        #     for k, wl in enumerate(wl_prf):
        #         if psf_format == 'pixel_based':
        #             enc += [find_aperture_radius(prfs[k],
        #                                          self.description['EnE'],
        #                                          1, 1, 1)]
        #         else:
        #             enc += [find_aperture_radius(prfs[k],
        #                                          self.description['EnE'],
        #                                          self.description['Fnum_x'][
        #                                              'value'],
        #                                          self.description['Fnum_y'][
        #                                              'value'],
        #                                          delta)]
        #     EncE_sol = interp1d(wl_prf, enc, assume_sorted=False)
        #     radius = EncE_sol(self.table['Wavelength'])
        else:
            radius = 1.22
        self.debug("radius : {}".format(radius))

        window_spatial_width = (
            2.0
            * radius
            * self.description["Fnum_y"]["value"]
            * self.table["Wavelength"]
            / delta
        ).to("")
        if "window_spatial_scale" in list(self.description.keys()):
            window_spatial_width *= self.description["window_spatial_scale"][
                "value"
            ]
            self.debug(
                "window spatial width scaled by {}".format(
                    self.description["window_spatial_scale"]["value"]
                )
            )
        if "window_spatial_pixel" in list(self.description.keys()):
            window_spatial_width = np.full(
                len(self.table["Wavelength"]),
                float(self.description["window_spatial_pixel"]["value"]),
            )
            self.debug(
                "window spatial width set to {}".format(
                    self.description["window_spatial_pixel"]["value"]
                )
            )
        if "window_spectral_pixel" in list(self.description.keys()):
            window_spectral_width = np.full(
                len(self.table["Wavelength"]),
                float(self.description["window_spectral_pixel"]["value"]),
            )
            self.debug(
                "window spectral width set to {}".format(
                    self.description["window_spectral_pixel"]["value"]
                )
            )
        window_size_px = window_spectral_width * window_spatial_width
        self._add_data_to_built("window_spectral_width", window_spectral_width)
        self._add_data_to_built("window_spatial_width", window_spatial_width)
        self._add_data_to_built("window_size_px", window_size_px)
        self.table["WindowSize"] = window_size_px
        self.debug("window size : {}".format(window_size_px))

    def propagate_target(self, target):
        out = QTable()
        wl = target.star.sed.wl_grid
        qe, transmission, wave_window = self._get_efficiency(wl, target)
        if "sky TR" in self.table.keys():
            out["foreground_transmission"] = self.table["sky TR"]
        if self.payload["optics"]["ForceChannelWlEdge"]["value"]:
            self.debug("force channel wl edge enabled")
            idx = np.logical_or(
                wl
                < self.description["wl_min"]["value"].to(
                    self.table["Wavelength"].unit
                ),
                wl
                > self.description["wl_max"]["value"].to(
                    self.table["Wavelength"].unit
                ),
            )
            transmission[idx] = wave_window[idx] = 0.0

        signal_density = CustomSignal(
            wl_grid=target.star.sed.wl_grid,
            data=(
                self.payload["optics"]["Atel"]["value"]
                * transmission
                * qe
                * target.star.sed.data
                * target.star.sed.wl_grid.to(u.m)
                / const.h
                / const.c
            ).to(1 / u.um / u.s)
            * u.count,
            data_unit=u.count / u.um / u.s,
        )
        self.debug("star signal density: {}".format(signal_density.data))

        # max signal in pixel
        gain_prf_data = self.built_instr["gain_prf_data"]
        gain_prf = interp1d(
            gain_prf_data["wl_grid"]["value"],
            gain_prf_data["data"]["value"],
            kind="cubic",
            assume_sorted=False,
        )

        # signal in spectral bin
        window_function = self._window_function(target.star.sed)
        self.debug("window function: {}".format(window_function))

        flux_density = wave_window * target.star.sed.data
        self.debug("star flux density: {}".format(flux_density))

        star_flux = np.trapz(
            flux_density * window_function, x=target.star.sed.wl_grid.to(u.um)
        ).to(u.W / u.m**2)
        self.debug("star flux : {}".format(star_flux))
        out["starFlux"] = star_flux

        star_signal = np.trapz(
            signal_density.data * window_function, x=target.star.sed.wl_grid
        ).to(u.count / u.s)
        self.debug("star signal : {}".format(star_signal))
        out["starSignal"] = star_signal
        out["star_signal_inAperture"] = star_signal

        star_signal_inPixel_density = signal_density
        star_signal_inPixel_density.spectral_rebin(
            self.built_instr["wl_pix_center"]
        )

        star_signal_inPixel = CountsPerSeconds(
            star_signal_inPixel_density.wl_grid,
            (
                star_signal_inPixel_density.data
                * self.built_instr["pixel_bandwidth"]
                * gain_prf(star_signal_inPixel_density.wl_grid)
            ).to(u.count / u.s),
        )
        starSignal_inPixel_max = (
            np.empty(self.table["Wavelength"].size, dtype=float)
            * star_signal_inPixel.data.unit
        )
        for k, (wld, wlu) in enumerate(
            zip(self.table["LeftBinEdge"], self.table["RightBinEdge"])
        ):
            idx = np.where(
                np.logical_and(
                    star_signal_inPixel.wl_grid > wld,
                    star_signal_inPixel.wl_grid < wlu,
                )
            )[0]
            starSignal_inPixel_max[k] = np.max(star_signal_inPixel.data[idx])
        out["star_MaxSignal_inPixel"] = starSignal_inPixel_max
        self.debug(
            "star signal in pixel MAX : {}".format(starSignal_inPixel_max)
        )
        return out



import logging
import unittest

from exorad.log import setLogLevel
from exorad.utils.version_control import VersionControl
from exorad.utils.version_control import VersionError

setLogLevel(logging.DEBUG)


class VersionControlTest(unittest.TestCase):

    def test_updated(self):
        VersionControl()

    def test_to_update(self):
        VersionControl(current_version='2.0.0')

    def test_error(self):
        with self.assertRaises(VersionError) as context:
            ver = VersionControl(current_version='2.0.0', force=True)
            if not ver.status_code:
                raise VersionError



import os
import xml.etree.ElementTree as ET
from collections import OrderedDict

from astropy import units as u
from astropy.io import ascii
from astropy.io.ascii.core import InconsistentTableError

from .task import Task
from exorad.utils.passVal import PassVal

compactString = lambda string: string.replace("\n", "").strip()


class LoadOptions(Task):
    """
    Reads the xml file with payload description and return an object with attributes related to the input data

    Parameters
    ----------
    filename: string
        input xml file location
    config_path: string (optional)
        on-run setting for ConfigPat. Default is None.

    Returns
    -------
    dict:
        parsed xml input file

    Raises
    ------
        IOError
            if the indicated file is not found or the format is not supported

    Examples
    --------
    >>> loadOptions = LoadOptions()
    >>> options = loadOptions(filename = 'path/to/file.xml')
    """

    def __init__(self):
        self.addTaskParam("filename", "input option file name")
        self.addTaskParam("config_path", "on-run setting for ConfigPath", None)

    def execute(self):
        self._filename = self.get_task_param("filename")
        self._config_path = self.get_task_param("config_path")
        self.__check_format__()
        root = self.__get_root__()
        self._dict = self.__parser__(root)
        # self.debug('loaded options: {}'.format(self._dict))
        self.set_output(self._dict)

    configPath = None

    def __check_format__(self):
        if not self._filename.endswith(".xml"):
            self.error("wrong input format")
            raise OSError

    def __get_root__(self):
        try:
            self.debug("input option file found %s" % self._filename)
            return ET.parse(self._filename).getroot()
        except OSError:
            self.error("No input option file found")
            raise OSError

    def __parser__(self, root):
        root_dict = {}

        for ch in root:
            retval = self.__parser__(ch)
            # parse all attributes
            for key in list(ch.attrib.keys()):
                retval[key] = ch.attrib[key]

            value = compactString(ch.text)
            if (value is not None) and (value != ""):
                try:
                    value = int(value)
                except ValueError:
                    if value.replace(".", ""):
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                if "unit" in retval:
                    unitName = retval["unit"]
                    if unitName == "dimensionless":
                        unitName = ""
                    value = value * u.Unit(unitName)
                if value == "True":
                    value = bool(True)
                if value == "False":
                    value = bool(False)
                if isinstance(value, str):
                    if "__ConfigPath__" in value:
                        value = value.replace(
                            "__ConfigPath__", self.configPath
                        )
                if ch.tag == "ConfigPath":
                    self.configPath = value
                if self._config_path:
                    self.configPath = self._config_path
                if ch.tag == "working_R":
                    PassVal.working_R = value
                    print(PassVal.working_R)
                # if isinstance(value, str):
                #     retval = value
                # else:
                retval["value"] = value

            if ch.tag in root_dict:
                # if an other instance of same tag exists, transform into a dict
                attr = root_dict[ch.tag]
                if isinstance(attr, OrderedDict):
                    attr[value] = retval
                else:
                    if not isinstance(attr, str):
                        dtmp = OrderedDict(
                            [(attr["value"], attr), (value, retval)]
                        )
                        root_dict[ch.tag] = dtmp
            else:
                # othewise, set new attr
                root_dict[ch.tag] = retval

        if "datafile" in root_dict:
            try:
                datatype_attr = root_dict["datatype"]
                datatype = datatype_attr["value"]
            except:
                datatype = "ecsv"

            attrValue = root_dict["datafile"]
            datafile = attrValue
            if "__ConfigPath__" in datafile["value"]:
                datafile = datafile["value"].replace(
                    "__ConfigPath__", self.configPath
                )
            else:
                datafile = datafile["value"]
            try:
                data = self.__read_datatable__(datafile, datatype)
                root_dict["data"] = data
            except OSError:
                self.error("Cannot read input file")
                raise OSError

        if "datadict" in root_dict:
            attrValue = root_dict["datadict"]
            datafile = attrValue
            datafile = datafile["value"].replace(
                "__ConfigPath__", self.configPath
            )
            try:
                data = self.__read_datadict__(datafile)
                root_dict["data"] = data
            except OSError:
                self.error("Cannot read input file")
                raise OSError
        if "config_dict" in root_dict:
            attrValue = root_dict["config_dict"]
            datafile = attrValue
            datafile = datafile["value"].replace(
                "__ConfigPath__", self.configPath
            )
            try:
                data = self.__read_datadict__(datafile)
                root_dict = data
                root_dict["config_dict"] = datafile
                self.debug(
                    "Configuration dictionary found in {}".format(datafile)
                )
            except OSError:
                self.error("Cannot read input file")
                raise OSError

        return root_dict

    def getopt(self):
        return self.__obj__

    def __read_datatable__(self, datafile, datatype):
        if datatype == "ecsv":
            try:
                data = ascii.read(
                    os.path.expanduser(datafile),
                    format="ecsv",
                    fill_values=[("#N/A", "0"), ("N/A", "0"), ("", "0")],
                )
            except InconsistentTableError:
                data = ascii.read(
                    os.path.expanduser(datafile),
                    format="csv",
                    fill_values=[("#N/A", "0"), ("N/A", "0"), ("", "0")],
                    delimiter=",",
                )
            return data

        else:
            self.error("inconsistent table")
            raise OSError

    def __read_datadict__(self, datadict):
        ext = os.path.splitext(datadict)[1]
        if ext in [".h5", ".hdf5"]:
            from exorad.output.hdf5 import load

            data = load(datadict)
            return data
        elif ext == ".pickle":
            import pickle

            f = open(datadict, "rb")
            data = pickle.load(f)
            f.close()
            return data
        else:
            self.error("inconsistent table")
            raise OSError
